from django_unicorn.components import UnicornView
from example.coffee.models import Flavor


class RowView(UnicornView):
    model: Flavor = None
    is_editing = False
    is_updated_by_child = False

    def edit(self):
        self.is_editing = True

    def cancel(self):
        self.is_editing = False

    def save(self):
        self.model.save()
        self.is_editing = False



from django.db.models import Model


class DbModel:
    def __init__(self, name: str, model_class: Model, *, defaults: dict | None = None):
        if defaults is None:
            defaults = {}
        self.name = name
        self.model_class = model_class
        self.defaults = defaults



import time
from uuid import uuid4

import pytest
from django.http import JsonResponse

from django_unicorn.errors import ComponentClassLoadError, ComponentModuleLoadError
from django_unicorn.views import message


def assert_json_error(response, error):
    assert isinstance(response, JsonResponse)
    assert response.status_code == 200
    assert response.json().get("error") == error


def post_json(client, data, url="/message/test-message"):
    return client.post(url, data, content_type="application/json")


def test_view_csrf(rf):
    request = rf.post("/message/test-message-csrf")
    response = message(request)
    assert response.status_code == 403


def test_message_component_name_is_none(client):
    response = client.post("/message")

    assert_json_error(response, "Missing component name in url")


def test_message_no_body(client):
    response = client.post("/message/test-message-no-body")

    assert_json_error(response, "Body could not be parsed")


def test_message_no_data(client):
    data = {}
    response = post_json(client, data)

    assert_json_error(response, "Invalid JSON body")


def test_message_no_checksum(client):
    data = {
        "data": {},
        "id": str(uuid4()),
        "epoch": time.time(),
    }
    response = post_json(client, data, url="/message/test-message-no-checksum")

    assert_json_error(response, "Missing checksum")


def test_message_bad_checksum(client):
    data = {
        "data": {},
        "checksum": "asdf",
        "id": str(uuid4()),
        "epoch": time.time(),
    }
    response = post_json(client, data, url="/message/test-message-bad-checksum")

    assert_json_error(response, "Checksum does not match")


def test_message_no_component_id(client):
    data = {
        "data": {},
        "checksum": "DVVk97cx",
        "epoch": time.time(),
    }
    response = post_json(client, data, url="/message/test-message-no-component-id")

    assert_json_error(response, "Missing component id")


def test_message_no_epoch(client):
    data = {"data": {}, "checksum": "DVVk97cx", "id": str(uuid4())}
    response = post_json(client, data, url="/message/test-message-no-epoch")

    assert_json_error(response, "Missing epoch")


def test_message_component_module_not_loaded(client):
    data = {
        "data": {},
        "checksum": "DVVk97cx",
        "id": str(uuid4()),
        "epoch": time.time(),
    }

    with pytest.raises(ComponentModuleLoadError) as e:
        post_json(client, data, url="/message/test-message-module-not-loaded")

    assert (
        e.exconly()
        == "django_unicorn.errors.ComponentModuleLoadError: The component module 'test_message_module_not_loaded' could not be loaded."  # noqa: E501
    )
    assert e.value.locations == [
        ("unicorn.components.test_message_module_not_loaded", "TestMessageModuleNotLoadedView"),
        ("components.test_message_module_not_loaded", "TestMessageModuleNotLoadedView"),
    ]


def test_message_component_class_not_loaded(client):
    data = {
        "data": {},
        "checksum": "DVVk97cx",
        "id": str(uuid4()),
        "epoch": time.time(),
    }

    with pytest.raises(ComponentClassLoadError) as e:
        post_json(
            client,
            data,
            url="/message/tests.views.fake_components.FakeComponentNotThere",
        )

    expected_exception_message = "django_unicorn.errors.ComponentClassLoadError: The component class \
'tests.views.fake_components.FakeComponentNotThere' could not be loaded: module 'tests.views.fake_components' \
has no attribute 'FakeComponentNotThere'"

    assert e.exconly() == expected_exception_message

    expected_locations = [
        ("tests.views.fake_components", "FakeComponentNotThere"),
        (
            "unicorn.components.tests.views.fake_components.FakeComponentNotThere",
            "FakecomponentnotthereView",
        ),
        (
            "components.tests.views.fake_components.FakeComponentNotThere",
            "FakecomponentnotthereView",
        ),
    ]
    assert e.value.locations == expected_locations


def test_message_component_class_with_attribute_error(client):
    data = {
        "data": {},
        "checksum": "DVVk97cx",
        "id": str(uuid4()),
        "epoch": time.time(),
    }

    with pytest.raises(ComponentClassLoadError) as e:
        post_json(
            client,
            data,
            url="/message/tests.views.fake_components.FakeComponentWithError",
        )

    assert e.value.__cause__


def test_message_component_with_dash(client):
    data = {
        "data": {},
        "checksum": "DVVk97cx",
        "id": str(uuid4()),
        "epoch": time.time(),
    }

    with pytest.raises(ComponentModuleLoadError) as e:
        post_json(client, data, url="/message/test-with-dash")

    assert (
        e.exconly()
        == "django_unicorn.errors.ComponentModuleLoadError: The component module 'test_with_dash' could not be loaded."
    )


def test_message_component_with_dot(client):
    data = {
        "data": {},
        "checksum": "DVVk97cx",
        "id": str(uuid4()),
        "epoch": time.time(),
    }

    with pytest.raises(ComponentClassLoadError) as e:
        post_json(client, data, url="/message/test.dot")

    expected = "django_unicorn.errors.ComponentClassLoadError: The component class 'test.dot' could not be loaded: \
module 'test' has no attribute 'dot'"
    actual = e.exconly()

    assert actual == expected



from django_unicorn.components import UnicornView
from django_unicorn.views.utils import _is_component_field_model_or_unicorn_field
from example.coffee.models import Flavor


class TypeHintView(UnicornView):
    model: Flavor | None = None


class ModelInstanceView(UnicornView):
    model = Flavor()


def test_type_hint():
    component = TypeHintView(component_name="asdf", component_id="test_type_hint")
    name = "model"
    actual = _is_component_field_model_or_unicorn_field(component, name)

    assert actual
    assert component.model is not None
    assert type(component.model) is Flavor


def test_model_instance():
    component = ModelInstanceView(component_name="asdf", component_id="test_model_instance")
    name = "model"
    actual = _is_component_field_model_or_unicorn_field(component, name)

    assert actual
    assert component.model is not None
    assert type(component.model) is Flavor



import logging
from datetime import timedelta
from decimal import Decimal
from functools import lru_cache
from types import MappingProxyType
from typing import Any

import orjson
from django.core.serializers import serialize
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import (
    DateField,
    DateTimeField,
    DurationField,
    Model,
    QuerySet,
    TimeField,
)
from django.utils.dateparse import (
    parse_date,
    parse_datetime,
    parse_duration,
    parse_time,
)
from django.utils.duration import duration_string

from django_unicorn.utils import is_int, is_non_string_sequence

try:
    from pydantic import BaseModel as PydanticBaseModel
except ImportError:
    PydanticBaseModel = None  # type: ignore


logger = logging.getLogger(__name__)

django_json_encoder = DjangoJSONEncoder()


class JSONDecodeError(Exception):
    pass


class InvalidFieldNameError(Exception):
    def __init__(self, field_name: str, data: dict | None = None):
        message = f"Cannot resolve '{field_name}'."

        if data:
            available = ", ".join(data.keys())
            message = f"{message} Choices are: {available}"

        super().__init__(message)


class InvalidFieldAttributeError(Exception):
    def __init__(self, field_name: str, field_attr: str, data: dict | None = None):
        message = f"Cannot resolve '{field_attr}'."

        if data:
            available = ", ".join(data[field_name].keys())
            message = f"{message} Choices on '{field_name}' are: {available}"

        super().__init__(message)


def _parse_field_values_from_string(model: Model) -> None:
    """
    Convert the model fields' value to match the field type if appropriate.

    This is mostly to deal with field string values that will get saved as a date-related field.
    """

    for field in model._meta.fields:
        val = getattr(model, field.attname)

        if not isinstance(val, str):
            continue

        if isinstance(field, DateTimeField):
            setattr(model, field.attname, parse_datetime(val))
        elif isinstance(field, TimeField):
            setattr(model, field.attname, parse_time(val))
        elif isinstance(field, DateField):
            setattr(model, field.attname, parse_date(val))
        elif isinstance(field, DurationField):
            setattr(model, field.attname, parse_duration(val))


def _get_many_to_many_field_related_names(model: Model) -> list[str]:
    """
    Get the many-to-many fields for a particular model. Returns either the automatically
    defined field name (i.e. something_set) or the related name.
    """

    # Use this internal method so that the fields can be cached
    @lru_cache(maxsize=128, typed=True)
    def _get_many_to_many_field_related_names_from_meta(meta):
        names = []

        for field in meta.get_fields():
            if field.is_relation and field.many_to_many:
                related_name = field.name

                if field.auto_created:
                    related_name = field.related_name or f"{field.name}_set"

                names.append(related_name)

        return names

    return _get_many_to_many_field_related_names_from_meta(model._meta)


def _get_m2m_field_serialized(model: Model, field_name) -> list:
    pks = []

    try:
        related_descriptor = getattr(model, field_name)

        # Get `pk` from `all` because it will re-use the cached data if the m-2-m field is prefetched
        # Using `values_list("pk", flat=True)` or `only()` won't use the cached prefetched values
        pks = [m.pk for m in related_descriptor.all()]
    except ValueError:
        # ValueError is thrown when the model doesn't have an id already set
        pass

    return pks


def _handle_inherited_models(model: Model, model_json: dict):
    """
    Handle if the model has a parent (i.e. the model is a subclass of another model).

    Subclassed model's fields don't get serialized
    (https://docs.djangoproject.com/en/stable/topics/serialization/#inherited-models)
    so those fields need to be retrieved manually.
    """

    if model._meta.get_parent_list():
        for field in model._meta.get_fields():
            if field.name not in model_json and hasattr(field, "primary_key") and not field.primary_key:
                if field.is_relation:
                    # We already serialized the m2m fields above, so we can skip them, but need to handle FKs
                    if not field.many_to_many:
                        foreign_key_field = getattr(model, field.name)
                        foreign_key_field_pk = getattr(
                            foreign_key_field,
                            "pk",
                            getattr(foreign_key_field, "id", None),
                        )
                        model_json[field.name] = foreign_key_field_pk
                else:
                    value = getattr(model, field.name)

                    # Explicitly handle `timedelta`, but use the DjangoJSONEncoder for everything else
                    if isinstance(value, timedelta):
                        value = duration_string(value)
                    else:
                        # Make sure the value is properly serialized
                        value = django_json_encoder.encode(value)

                        # The DjangoJSONEncoder has extra double-quotes for strings so remove them
                        if isinstance(value, str) and value.startswith('"') and value.endswith('"'):
                            value = value[1:-1]

                    model_json[field.name] = value


def _get_model_dict(model: Model) -> dict:
    """
    Serializes Django models. Uses the built-in Django JSON serializer, but moves the data around to
    remove some unnecessary information and make the structure more compact.
    """

    _parse_field_values_from_string(model)

    # Django's `serialize` method always returns a string of an array
    serialized_model = serialize("json", [model])

    try:
        # Django's `serialize` method always returns a string of an array,
        # so remove the brackets from the resulting string
        model_json = orjson.loads(serialized_model[1:-1])
    except orjson.JSONDecodeError:
        # Convert the string into a dictionary and grab the `pk`
        model_json = orjson.loads(serialized_model)[0]
    model_pk = model_json.get("pk")

    # Shuffle around the serialized pieces to condense the size of the payload
    model_json = model_json.get("fields")
    model_json["pk"] = model_pk

    # Set `pk` for models that subclass another model which only have `id` set
    if not model_pk:
        model_json["pk"] = model.pk or model.id  # type: ignore

    # Add in m2m fields
    m2m_field_names = _get_many_to_many_field_related_names(model)

    for m2m_field_name in m2m_field_names:
        model_json[m2m_field_name] = _get_m2m_field_serialized(model, m2m_field_name)

    _handle_inherited_models(model, model_json)

    return model_json


def _json_serializer(obj):
    """
    Handle the objects that the `orjson` deserializer can't handle automatically.

    The types handled by `orjson` by default: dataclass, datetime, enum, float, int, numpy, str, uuid.
    The types handled in this class: Django Model, Django QuerySet, Decimal, or any object with `to_json` method.

    TODO: Investigate other ways to serialize objects automatically.
    e.g. Using DRF serializer: https://www.django-rest-framework.org/api-guide/serializers/#serializing-objects
    """
    from django_unicorn.components import UnicornView  # noqa: PLC0415

    try:
        if isinstance(obj, UnicornView):
            return {
                "name": obj.component_name,
                "id": obj.component_id,
                "key": obj.component_key,
            }
        elif isinstance(obj, Model):
            return _get_model_dict(obj)
        elif isinstance(obj, QuerySet):
            queryset_json = []

            for model in obj:
                if obj.query.values_select and isinstance(model, dict):
                    # If the queryset was created with values it's already a dictionary
                    model_json = model
                else:
                    model_json = _get_model_dict(model)

                queryset_json.append(model_json)

            return queryset_json
        elif PydanticBaseModel and isinstance(obj, PydanticBaseModel):  # type: ignore
            return obj.dict()
        elif isinstance(obj, Decimal):
            return str(obj)
        elif isinstance(obj, MappingProxyType):
            # Return a regular dict for `mappingproxy`
            return obj.copy()
        elif hasattr(obj, "to_json"):
            return obj.to_json()
    except Exception as e:
        # Log this because the `TypeError` and resulting stacktrace lacks context
        logger.exception(e)

    raise TypeError


def _fix_floats(current: dict, data: dict | None = None, paths: list | None = None) -> None:
    """
    Recursively change any Python floats to a string so that JavaScript
    won't convert the float to an integer when deserializing.

    Params:
        current: Dictionary in which to check for and fix floats.
    """

    if data is None:
        data = current

    if paths is None:
        paths = []

    if isinstance(current, dict):
        for key, val in current.items():
            paths.append(key)
            _fix_floats(val, data, paths=paths)
            paths.pop()
    elif isinstance(current, list):
        for idx, item in enumerate(current):
            paths.append(idx)
            _fix_floats(item, data, paths=paths)
            paths.pop()
    elif isinstance(current, float):
        _piece = data

        for idx, path in enumerate(paths):
            if idx == len(paths) - 1:
                # `path` can be a dictionary key or list index,
                # but in either instance it is set the same way
                _piece[path] = str(current)
            else:
                _piece = _piece[path]


def _sort_dict(data: dict) -> dict:
    """
    Recursively sort the dictionary keys so that JavaScript won't change the order
    and change the generated checksum.

    Params:
        data: Dictionary to sort.
    """

    if not isinstance(data, dict):
        return data

    items = [
        [k, v]
        for k, v in sorted(
            data.items(),
            key=lambda item: item[0] if not is_int(item[0]) else int(item[0]),
        )
    ]

    for item in items:
        if isinstance(item[1], dict):
            item[1] = _sort_dict(item[1])

    return dict(items)


def _exclude_field_attributes(dict_data: dict[Any, Any], exclude_field_attributes: tuple[str] | None = None) -> None:
    """
    Remove the field attribute from `dict_data`. Handles nested attributes with a dot.

    Example:
    _exclude_field_attributes({"1": {"2": {"3": "4"}}}, ("1.2.3",)) == {"1": {"2": {}}}
    """

    if exclude_field_attributes:
        for field in exclude_field_attributes:
            field_splits = field.split(".")
            nested_attribute_split_count = 2

            if len(field_splits) > nested_attribute_split_count:
                next_attribute_index = field.index(".") + 1
                remaining_field_attributes = field[next_attribute_index:]
                remaining_dict_data = dict_data[field_splits[0]]

                return _exclude_field_attributes(remaining_dict_data, (remaining_field_attributes,))
            elif len(field_splits) == nested_attribute_split_count:
                (field_name, field_attr) = field_splits

                if field_name not in dict_data:
                    raise InvalidFieldNameError(field_name=field_name, data=dict_data)

                if dict_data[field_name] is not None:
                    if field_attr not in dict_data[field_name]:
                        raise InvalidFieldAttributeError(field_name=field_name, field_attr=field_attr, data=dict_data)

                    del dict_data[field_name][field_attr]


@lru_cache(maxsize=128, typed=True)
def _dumps(
    serialized_data: bytes,
    *,
    fix_floats: bool = True,
    exclude_field_attributes: tuple[str] | None = None,
    sort_dict: bool = True,
) -> dict:
    """
    Dump serialized data with custom massaging.

    Features:
       - fix floats
       - remove specific keys as needed
       - sort dictionary
    """

    data = orjson.loads(serialized_data)

    if fix_floats:
        _fix_floats(data)

    if exclude_field_attributes:
        # Excluding field attributes needs to de-serialize and then serialize again to
        # handle complex objects
        _exclude_field_attributes(data, exclude_field_attributes)

    if sort_dict:
        # Sort dictionary manually because stringified integers don't get sorted
        # correctly with `orjson.OPT_SORT_KEYS` and JavaScript will sort the keys
        # as if they are integers
        data = _sort_dict(data)

    return data


def dumps(
    data: Any,
    *,
    fix_floats: bool = True,
    exclude_field_attributes: tuple[str, ...] | None = None,
    sort_dict: bool = True,
) -> str:
    """
    Converts the passed-in dictionary to a string representation.

    Handles the following objects: dataclass, datetime, enum, float, int, numpy, str, uuid,
    Django Model, Django QuerySet, Pydantic models (`PydanticBaseModel`), any object with `to_json` method.

    Args:
        param fix_floats: Whether any floats should be converted to strings. Defaults to `True`,
            but will be faster without it.
        param exclude_field_attributes: Tuple of strings with field attributes to remove, i.e. "1.2"
            to remove the key `2` from `{"1": {"2": "3"}}`
        param sort_dict: Whether the `dict` should be sorted. Defaults to `True`, but
            will be faster without it.

    Returns a `str` instead of `bytes` (which deviates from `orjson.dumps`), but seems more useful.
    """

    if exclude_field_attributes is not None and not is_non_string_sequence(exclude_field_attributes):
        raise AssertionError("exclude_field_attributes type needs to be a sequence")

    # Call `dumps` to make sure that complex objects are serialized correctly
    serialized_data = orjson.dumps(data, default=_json_serializer)

    data = _dumps(
        serialized_data,
        fix_floats=fix_floats,
        exclude_field_attributes=exclude_field_attributes,
        sort_dict=sort_dict,
    )

    serialized_data = orjson.dumps(data)

    return serialized_data.decode("utf-8")


def loads(string: str) -> dict:
    """
    Converts a string representation to dictionary.
    """

    try:
        return orjson.loads(string)
    except orjson.JSONDecodeError as e:
        raise JSONDecodeError from e


def model_value(model: Model, *fields: str):
    """
    Serializes a model into a dictionary with the fields as specified in the `fields` argument.
    """

    model_data = {}
    model_dict = _get_model_dict(model)

    if not fields:
        return model_dict

    for field in fields:
        if field in model_dict:
            model_data[field] = model_dict[field]

    return model_data



import pickle
import sys
from unittest.mock import MagicMock

from bs4 import BeautifulSoup
from django.test import RequestFactory

from django_unicorn.components import UnicornView
from django_unicorn.components.unicorn_template_response import UnicornTemplateResponse


class RecursionComponent(UnicornView):
    template_name = "templates/test_component.html"
    pass


def test_child_component_cleanup_prevents_recursion_error():
    # 1. Create child component and simulate it having a large _json_tag (as if rendered)
    child = RecursionComponent(component_name="child", component_id="child_id")

    # Create a deep DOM to ensure recursion error would happen if not cleaned up
    depth = 2000
    html_content = "<div>" + ("<span>" * depth) + "Hello" + ("</span>" * depth) + "</div>"
    soup = BeautifulSoup(html_content, "html.parser")
    child._json_tag = soup.find("div")  # Simulate the tag attached during child render
    child._init_script = "Unicorn.componentInit({});"

    # 2. Create parent component and link child
    parent = RecursionComponent(component_name="parent", component_id="parent_id")
    parent.children.append(child)

    # 3. Setup UnicornTemplateResponse for parent
    request = RequestFactory().get("/")
    template = MagicMock()
    template.render.return_value = "<html><body><div unicorn:view>Parent Content</div></body></html>"

    response = UnicornTemplateResponse(
        template=template,
        request=request,
        component=parent,
        init_js=True,  # Important to trigger the logic that collects json_tags
    )

    # Increase recursion limit to allow BS4 to process the deep tree without crashing during render
    # We want to verify that pickling works AFTER the tag is removed, not that BS4 crashes.
    original_recursion_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(10000)

    try:
        # 4. Render the parent
        # This should trigger the loop that collects children's json_tags and DELETES them
        response.render()

        # 5. Verify _json_tag is deleted from child
        assert not hasattr(child, "_json_tag"), "child._json_tag should have been deleted"

        # 6. Verify pickling parent works
        # If _json_tag was arguably still there, this might fail or pass depending on limit.
        # But since we verified it IS deleted, this confirms the object state is clean.
        pickle.dumps(parent)

    finally:
        sys.setrecursionlimit(original_recursion_limit)



from django_unicorn.components import UnicornView
from django_unicorn.typing import QuerySetType
from typing import Optional, cast, Any

from example.coffee.models import Flavor, Taste


class ModelsView(UnicornView):
    flavor: Optional[Flavor] = None
    flavors: QuerySetType[Flavor] = cast(QuerySetType[Flavor], Flavor.objects.none())

    def mount(self):
        self.flavor = Flavor()
        self.refresh_flavors()

    def refresh_flavors(self):
        self.flavors = cast(QuerySetType[Flavor], Flavor.objects.all().order_by("-id")[:2])

    def save_flavor(self):
        if self.flavor:
            self.flavor.save()
        self.flavor = Flavor()
        self.refresh_flavors()

    def save(self, flavor_idx: int):
        flavor_data = self.flavors[flavor_idx]
        flavor_data.save()

    def delete(self, flavor_to_delete: Flavor):
        flavor_to_delete.delete()
        self.refresh_flavors()

    def available_flavors(self):
        return Flavor.objects.all()

    def available_tastes(self):
        return Taste.objects.all()

    def model_typehint(self, flavor: Flavor):
        print(flavor)

    class Meta:
        javascript_exclude = ("flavor.taste_set",)



import time

import orjson
import shortuuid

from django_unicorn.utils import generate_checksum


def _post_message_and_get_body(client, message):
    response = client.post(
        "/message/tests.views.fake_components.FakeComponent",
        message,
        content_type="application/json",
    )

    body = orjson.loads(response.content)
    return body


def test_message_toggle(client):
    data = {"check": False}
    message = {
        "actionQueue": [
            {"type": "callMethod", "payload": {"name": "$toggle('check')"}},
        ],
        "data": data,
        "checksum": generate_checksum(str(data)),
        "id": shortuuid.uuid()[:8],
        "epoch": time.time(),
    }

    body = _post_message_and_get_body(client, message)

    assert not body["errors"]
    assert body["data"]["check"] is True


def test_message_nested_toggle(client):
    data = {"nested": {"check": False}}
    message = {
        "actionQueue": [
            {"type": "callMethod", "payload": {"name": "$toggle('nested.check')"}},
        ],
        "data": data,
        "checksum": generate_checksum(str(data)),
        "id": shortuuid.uuid()[:8],
        "epoch": time.time(),
    }

    body = _post_message_and_get_body(client, message)

    assert not body["errors"]
    assert body["data"]["nested"]["check"] is True



from django_unicorn.components import UnicornField, UnicornView
from django_unicorn.views.utils import set_property_from_data


class NestedPropertyOne(UnicornField):
    name = "nested_property_one"


class PropertyOne(UnicornField):
    nested_property_one = NestedPropertyOne()
    name = "property_one"


class NestedPropertyView(UnicornView):
    property_one = PropertyOne()
    name = "property_view"


def test_set_property_from_data_unicorn_field():
    component = NestedPropertyView(component_name="test", component_id="test_set_property_from_data_unicorn_field")
    assert "property_one" == component.property_one.name

    data = {"name": "property_one_updated"}
    set_property_from_data(component, "property_one", data)

    assert "property_one_updated" == component.property_one.name


def test_set_property_from_data_nested_unicorn_field():
    component = NestedPropertyView(
        component_name="test", component_id="test_set_property_from_data_nested_unicorn_field"
    )
    assert "nested_property_one" == component.property_one.nested_property_one.name

    data = {"nested_property_one": {"name": "nested_property_one_updated"}}
    set_property_from_data(component, "property_one", data)

    assert "nested_property_one_updated" == component.property_one.nested_property_one.name



import logging
from dataclasses import is_dataclass
from typing import Any, Union, cast

from django.db.models import Model

from django_unicorn.components import UnicornField, UnicornView
from django_unicorn.decorators import timed
from django_unicorn.typer import (
    cast_value,
    create_queryset,
    get_args,
    get_origin,
    get_type_hints,
    is_queryset,
)

try:
    from types import UnionType
except ImportError:
    UnionType = Union  # type: ignore


logger = logging.getLogger(__name__)


@timed
def set_property_from_data(
    component_or_field: UnicornView | UnicornField | Model,
    name: str,
    value: Any,
) -> None:
    """
    Sets properties on the component based on passed-in data.
    """

    try:
        if not hasattr(component_or_field, name):
            return
    except ValueError:
        # Treat ValueError the same as a missing field because trying to access a many-to-many
        # field before the model's pk will throw this exception
        return

    field = getattr(component_or_field, name)
    component_field_is_model_or_unicorn_field = _is_component_field_model_or_unicorn_field(component_or_field, name)

    # UnicornField and Models are always a dictionary (can be nested)
    if component_field_is_model_or_unicorn_field:
        # Re-get the field since it might have been set in `_is_component_field_model_or_unicorn_field`
        field = getattr(component_or_field, name)

        type_hints = get_type_hints(component_or_field)

        if isinstance(value, dict):
            for key in value.keys():
                key_value = value[key]
                set_property_from_data(field, key, key_value)
    elif hasattr(field, "related_val"):
        # Use `related_val` to check for many-to-many
        field.set(value)
    else:
        # TODO: Call django_unicorn.typing.cast_attribute_value?
        type_hints = get_type_hints(component_or_field)
        type_hint = type_hints.get(name)

        if is_queryset(field, type_hint, value):
            value = create_queryset(field, type_hint, value)
        elif type_hint:
            if is_dataclass(type_hint):
                value = type_hint(**value)
            else:
                try:
                    value = cast_value(type_hint, value)
                except TypeError:
                    # Ignore this exception because some type-hints can't be instantiated like this (e.g. `List[]`)
                    pass

        if hasattr(component_or_field, "_set_property"):
            # Can assume that `component_or_field` is a component
            cast(UnicornView, component_or_field)._set_property(
                name, value, call_updating_method=True, call_updated_method=False
            )
        else:
            setattr(component_or_field, name, value)


@timed
def _is_component_field_model_or_unicorn_field(
    component_or_field: UnicornView | UnicornField | Model,
    name: str,
) -> bool:
    """
    Determines whether a component's field is a Django `Model` or `UnicornField` either
    by checking the field's instance or inspecting the type hints.

    One side-effect is that the field will be instantiated if it is currently `None` and
    the type hint is available.

    Args:
        component: `UnicornView` to check.
        name: Name of the field.

    Returns:
        Whether the field is a Django `Model` or `UnicornField`.
    """
    field = getattr(component_or_field, name)

    if isinstance(field, Model | UnicornField):
        return True

    is_subclass_of_model = False
    is_subclass_of_unicorn_field = False
    component_type_hints = {}

    try:
        component_type_hints = get_type_hints(component_or_field)

        if name in component_type_hints:
            type_hint = component_type_hints[name]

            if get_origin(type_hint) is Union or get_origin(type_hint) is UnionType:
                for arg in get_args(type_hint):
                    if arg is not type(None):
                        type_hint = arg
                        break

            is_subclass_of_model = issubclass(type_hint, Model)

            if not is_subclass_of_model:
                is_subclass_of_unicorn_field = issubclass(type_hint, UnicornField)

            # Construct a new class if the field is None and there is a type hint available
            if field is None:
                if is_subclass_of_model or is_subclass_of_unicorn_field:
                    field = type_hint()
                    setattr(component_or_field, name, field)
    except TypeError:
        pass

    return is_subclass_of_model or is_subclass_of_unicorn_field



import pytest
from django.template.backends.django import Template

from django_unicorn.utils import (
    create_template,
    generate_checksum,
    get_method_arguments,
    is_non_string_sequence,
    sanitize_html,
)


def test_generate_checksum_bytes(settings):
    settings.SECRET_KEY = "asdf"

    expected = "TfxFqcQL"
    actual = generate_checksum(b'{"name": "test"}')

    assert expected == actual


def test_generate_checksum_str(settings):
    settings.SECRET_KEY = "asdf"

    expected = "TfxFqcQL"
    actual = generate_checksum('{"name": "test"}')

    assert expected == actual


def test_generate_checksum_dict(settings):
    settings.SECRET_KEY = "asdf"

    # This is different than the above because `str(dict)` turns `{"name": "test"}` into `"{'name': 'test'}"`
    expected = "JaV4PeA6"
    actual = generate_checksum({"name": "test"})

    assert expected == actual


def test_generate_checksum_invalid(settings):
    settings.SECRET_KEY = "asdf"

    with pytest.raises(TypeError) as e:
        generate_checksum([])  # type: ignore

    assert e.exconly() == "TypeError: Invalid type: <class 'list'>"


def test_get_method_arguments():
    def test_func(input_str):
        return input_str

    expected = [
        "input_str",
    ]
    actual = get_method_arguments(test_func)
    assert actual == expected


def test_get_method_arguments_with_type_annotation():
    def test_func(input_str: str):
        return input_str

    expected = [
        "input_str",
    ]
    actual = get_method_arguments(test_func)
    assert actual == expected


def test_sanitize_html():
    expected = '{"id":"abcd123","name":"text-inputs","key":"asdf","data":{"name":"World","testing_thing":"Whatever \\u003C/script\\u003E \\u003Cscript\\u003Ealert(\'uh oh\')\\u003C/script\\u003E"},"calls":[],"hash":"hjkl"}'  # noqa: E501
    data = '{"id":"abcd123","name":"text-inputs","key":"asdf","data":{"name":"World","testing_thing":"Whatever </script> <script>alert(\'uh oh\')</script>"},"calls":[],"hash":"hjkl"}'  # noqa: E501
    actual = sanitize_html(data)
    assert actual == expected


def test_is_non_string_sequence_list():
    assert is_non_string_sequence(
        [
            "",
        ]
    )


def test_is_non_string_sequence_tuple():
    assert is_non_string_sequence(("",))


def test_is_non_string_sequence_set():
    assert is_non_string_sequence(
        {
            "",
        }
    )


def test_is_non_string_sequence_string():
    assert not is_non_string_sequence("")


def test_is_non_string_sequence_bytes():
    assert not is_non_string_sequence(b"")


def test_create_template_str():
    actual = create_template("<div>test string template</div>")

    assert actual
    assert isinstance(actual, Template)


def test_create_template_callable():
    def _get_template():
        return "<div>test callable template</div>"

    actual = create_template(_get_template)

    assert actual
    assert isinstance(actual, Template)



from django.db.models import Model
from django.db.models.fields import CharField

from django_unicorn.components import UnicornView
from django_unicorn.views.utils import set_property_from_data


class FakeModel(Model):
    name = CharField(max_length=255)

    class Meta:
        app_label = "www"


class ModelPropertyView(UnicornView):
    model = FakeModel(name="fake_model")


def test_set_property_from_data_model():
    component = ModelPropertyView(component_name="test", component_id="test_set_property_from_data_model")
    assert "fake_model" == component.model.name

    set_property_from_data(component, "model", {"name": "fake_model_updated"})

    assert "fake_model_updated" == component.model.name



import types

import orjson
import pytest
from tests.views.fake_components import (
    FakeAuthenticationComponent,
    FakeValidationComponent,
)

from django_unicorn.components import UnicornView
from django_unicorn.serializer import InvalidFieldNameError


class ExampleComponent(UnicornView):
    name = "World"

    def get_name(self):
        return "World"


@pytest.fixture()
def component():
    return ExampleComponent(component_id="asdf1234", component_name="example")


def test_init_with_template_name():
    class TestComponent(UnicornView):
        template_name = "unicorn/test.html"

    component = TestComponent(component_id="test_init_with_template_name", component_name="hello-world")
    assert component.template_name == "unicorn/test.html"


def test_init_with_get_template_names():
    class TestComponent(UnicornView):
        def get_template_names(self):
            return []

    component = TestComponent(component_id="test_init_with_get_template_names", component_name="hello-world")
    assert component.template_name is None


def test_init_attribute_names_cache(component):
    attribute_names_cache = component._attribute_names_cache
    assert len(attribute_names_cache) == 1
    assert attribute_names_cache[0] == "name"


def test_init_attribute_names(component):
    attribute_names = component._attribute_names()
    assert len(attribute_names) == 1
    assert attribute_names[0] == "name"


def test_init_attributes(component):
    attributes = component._attributes()
    assert len(attributes) == 1
    assert attributes["name"] == "World"


def test_init_properties():
    class TestComponent(UnicornView):
        @property
        def name(self):
            return "World"

    component = TestComponent(component_id="test_init_properties", component_name="hello-world")
    attributes = component._attributes()
    assert len(attributes) == 1
    assert attributes["name"] == "World"


def test_init_methods_cache(component):
    assert len(component._methods_cache) == 1


def test_init_methods(component):
    methods = component._methods()
    assert len(methods) == 1
    assert methods["get_name"]() == "World"


def test_get_frontend_context_variables_javascript_exclude(component):
    class Meta:
        javascript_exclude = ("name",)

    component.Meta = Meta

    frontend_context_variables = component.get_frontend_context_variables()
    frontend_context_variables_dict = orjson.loads(frontend_context_variables)
    assert len(frontend_context_variables_dict) == 0
    assert "name" not in frontend_context_variables_dict


def test_get_frontend_context_variables_javascript_exclude_invalid_field(component):
    class Meta:
        javascript_exclude = ("blob",)

    component.Meta = Meta

    with pytest.raises(InvalidFieldNameError):
        component.get_frontend_context_variables()


def test_get_frontend_context_variables_exclude_field(component):
    # Use internal class prevent class cache from causing issues
    class ExampleComponentWithMetaExclude(UnicornView):
        name = "world"

        class Meta:
            exclude = ("name",)

    component = ExampleComponentWithMetaExclude(
        component_id="test_get_frontend_context_variables_exclude_field", component_name="example"
    )

    frontend_context_variables = component.get_frontend_context_variables()
    frontend_context_variables_dict = orjson.loads(frontend_context_variables)
    assert len(frontend_context_variables_dict) == 0
    assert "name" not in frontend_context_variables_dict


def test_get_frontend_context_variables_exclude_field_invalid_type():
    # Use internal class prevent class cache from causing issues
    class ExampleComponentWithMetaInvalidExclude(UnicornView):
        name = "world"

        class Meta:
            exclude = ""

    with pytest.raises(AssertionError):
        ExampleComponentWithMetaInvalidExclude(
            component_id="test_get_frontend_context_variables_exclude_field_invalid_type", component_name="example"
        )


def test_get_frontend_context_variables_exclude_invalid_field():
    # Use internal class prevent class cache from causing issues
    class ExampleComponentWithMetaInvalidExclude(UnicornView):
        pass

        class Meta:
            exclude = ("blob",)

    with pytest.raises(InvalidFieldNameError):
        ExampleComponentWithMetaInvalidExclude(
            component_id="test_get_frontend_context_variables_exclude_invalid_field", component_name="example"
        )


def test_get_frontend_context_variables(component):
    frontend_context_variables = component.get_frontend_context_variables()
    frontend_context_variables_dict = orjson.loads(frontend_context_variables)
    assert len(frontend_context_variables_dict) == 1
    assert frontend_context_variables_dict.get("name") == "World"


def test_get_context_data(component):
    context_data = component.get_context_data()
    assert len(context_data) == 4  # `unicorn` and `view` are added to context data by default
    assert context_data.get("name") == "World"
    assert isinstance(context_data.get("get_name"), types.MethodType)


def test_get_context_data_component():
    class TestComponent(UnicornView):
        pass

    component = TestComponent(component_id="test_get_context_data_component", component_name="hello-world")
    actual = component.get_context_data()

    assert actual["unicorn"]
    assert actual["unicorn"]["component"] == component


def test_get_context_data_component_id():
    class TestComponent(UnicornView):
        pass

    component = TestComponent(component_id="test_get_context_data_component_id", component_name="hello-world")
    actual = component.get_context_data()

    assert actual["unicorn"]
    assert actual["unicorn"]["component_id"] == "test_get_context_data_component_id"


def test_get_context_data_component_name():
    class TestComponent(UnicornView):
        pass

    component = TestComponent(component_id="test_get_context_data_component_name", component_name="hello-world")
    actual = component.get_context_data()

    assert actual["unicorn"]
    assert actual["unicorn"]["component_name"] == "hello-world"


def test_get_context_data_component_key():
    class TestComponent(UnicornView):
        pass

    component = TestComponent(
        component_id="test_get_context_data_component_key",
        component_name="hello-world",
        component_key="key-key-key",
    )
    actual = component.get_context_data()

    assert actual["unicorn"]
    assert actual["unicorn"]["component_key"] == "key-key-key"


def test_is_public(component):
    assert component._is_public("test_name")


def test_is_public_protected(component):
    assert component._is_public("_test_name") is False


def test_is_public_http_method_names(component):
    assert component._is_public("http_method_names") is False


def test_meta_javascript_exclude():
    class TestComponent(UnicornView):
        name = "World"

        class Meta:
            javascript_exclude = ("name",)

    component = TestComponent(component_id="test_meta_javascript_exclude", component_name="hello-world")
    assert "name" not in component.get_frontend_context_variables()
    assert "name" in component.get_context_data()


def test_meta_javascript_exclude_nested_with_tuple():
    class TestComponent(UnicornView):
        name = {"Universe": {"World": "Earth"}}  # noqa: RUF012

        class Meta:
            javascript_exclude = ("name.Universe",)

    expected = '{"name":{}}'
    component = TestComponent(
        component_id="test_meta_javascript_exclude_nested_with_tuple", component_name="hello-world"
    )
    assert expected == component.get_frontend_context_variables()


def test_meta_javascript_exclude_nested_multiple_with_spaces():
    class TestComponent(UnicornView):
        name = {"Universe": {"World": "Earth"}}  # noqa: RUF012
        another = {"Neutral Milk Hotel": {"album": {"On Avery Island": 1996}}}  # noqa: RUF012

        class Meta:
            javascript_exclude = ("another.Neutral Milk Hotel.album",)

    expected = '{"another":{"Neutral Milk Hotel":{}},"name":{"Universe":{"World":"Earth"}}}'
    component = TestComponent(
        component_id="test_meta_javascript_exclude_nested_multiple_with_spaces", component_name="hello-world"
    )
    actual = component.get_frontend_context_variables()
    assert expected == actual


def test_meta_javascript_exclude_nested_with_list():
    class TestComponent(UnicornView):
        name = {"Universe": {"World": "Earth"}}  # noqa: RUF012

        class Meta:
            javascript_exclude = [  # noqa: RUF012
                "name.Universe",
            ]

    expected = '{"name":{}}'
    component = TestComponent(
        component_id="test_meta_javascript_exclude_nested_with_list", component_name="hello-world"
    )
    assert expected == component.get_frontend_context_variables()


def test_meta_exclude():
    class TestComponent(UnicornView):
        name = "World"

        class Meta:
            exclude = ("name",)

    component = TestComponent(component_id="test_meta_exclude", component_name="hello-world")
    assert "name" not in component.get_frontend_context_variables()
    assert "name" not in component.get_context_data()


def test_get_frontend_context_variables_form_with_boolean_field(component):
    """
    Form classes with BooleanField and CheckboxInput widget set the bool values to `None`
    without an explicit fix.
    """

    component = FakeValidationComponent(
        component_id="test_get_frontend_context_variables_form_with_boolean_field", component_name="example"
    )

    frontend_context_variables = component.get_frontend_context_variables()
    frontend_context_variables_dict = orjson.loads(frontend_context_variables)

    assert frontend_context_variables_dict.get("permanent") is not None


def test_get_frontend_context_variables_authentication_form(component):
    """
    `AuthenticationForm` have `request` as the first argument so `form_class` should
    init with data as a kwarg.
    """

    component = FakeAuthenticationComponent(
        component_id="test_get_frontend_context_variables_authentication_form", component_name="example"
    )

    component.get_frontend_context_variables()



from django_unicorn.components.mixins import ModelValueMixin
from django_unicorn.components.unicorn_view import UnicornField, UnicornView
from django_unicorn.components.updaters import HashUpdate, LocationUpdate, PollUpdate
from django_unicorn.typing import QuerySetType

__all__ = [
    "HashUpdate",
    "LocationUpdate",
    "ModelValueMixin",
    "PollUpdate",
    "QuerySetType",
    "UnicornField",
    "UnicornView",
]



import os


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = "p6b$i%36e_xg%*ok+55@uc(h9)#g+2fj#p%7g#-@y8s6+10q#7"
DEBUG = True
ALLOWED_HOSTS = ["localhost"]


INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # Internal apps
    "example.books",
    "example.coffee",
    "example.www",
    # Include django-unicorn package
    "django_unicorn",
    # Include internal unicorn app
    "example.unicorn",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "example.project.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "example.project.wsgi.application"


CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "unique-snowflake",
    }
}
# CACHES["default"] = {
#     "BACKEND": "django_redis.cache.RedisCache",
#     "LOCATION": "redis://localhost:6379",
#     "OPTIONS": {"CLIENT_CLASS": "django_redis.client.DefaultClient"},
# }

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}


AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_L10N = True
USE_TZ = True


STATIC_URL = "/static/"
STATICFILES_DIRS = (os.path.join(BASE_DIR, "static"),)
STATIC_ROOT = os.path.join(BASE_DIR, "staticfiles")


DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

UNICORN = {
    "SERIAL": {
        "ENABLED": True,
        "TIMEOUT": 5,
    },
    "CACHE_ALIAS": "default",
    "MORPHER": {
        # "NAME": "alpine",
        "NAME": "morphdom",
        "RELOAD_SCRIPT_ELEMENTS": True,
    },
    "MINIFIED": False,
}


LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
        },
        "null": {
            "class": "logging.NullHandler",
        },
    },
    "root": {
        "handlers": ["console"],
        "level": "WARNING",
    },
    "loggers": {
        "django": {
            "handlers": ["console"],
            "level": os.getenv("DJANGO_LOG_LEVEL", "INFO"),
            "propagate": False,
        },
        "django.server": {
            "handlers": ["null"],
            "level": "INFO",
            "propagate": False,
        },
        "profile": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
        "django_unicorn": {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
}



from django_unicorn.components import UnicornView


class Step1View(UnicornView):
    name: str
    email: str

    def mount(self):
        self.name = "Test"
        self.email = "test@example.com"

    def noop(self):
        print("no-op")



from django_unicorn.components import UnicornView


class Step2View(UnicornView):
    address: str
    city: str
    state: str
    zip_code: str

    def mount(self):
        self.address = "123 Main St"
        self.city = "Anytown"
        self.state = "CA"
        self.zip_code = "12345"



import uuid

from django.db import models


class Flavor(models.Model):
    name = models.CharField(max_length=255)
    label = models.CharField(max_length=255)
    parent = models.ForeignKey("self", blank=True, null=True, on_delete=models.SET_NULL)
    float_value = models.FloatField(blank=True, null=True)
    decimal_value = models.DecimalField(blank=True, null=True, max_digits=10, decimal_places=2)
    uuid = models.UUIDField(default=uuid.uuid4)
    datetime = models.DateTimeField(blank=True, null=True)
    date = models.DateField(blank=True, null=True)
    time = models.TimeField(blank=True, null=True)
    duration = models.DurationField(blank=True, null=True)

    def __str__(self):
        return self.name


class Favorite(models.Model):
    is_favorite = models.BooleanField(default=False)
    flavor = models.OneToOneField(Flavor, on_delete=models.CASCADE)


class Taste(models.Model):
    name = models.CharField(max_length=255)
    flavor = models.ManyToManyField(Flavor)


class Origin(models.Model):
    name = models.CharField(max_length=255)
    flavor = models.ManyToManyField(Flavor, related_name="origins")


class NewFlavor(Flavor):
    new_name = models.CharField(max_length=255)



import re

import pytest
from django.template import Context
from django.template.base import Parser, Token, TokenType

from django_unicorn.components import UnicornView
from django_unicorn.errors import ComponentNotValidError
from django_unicorn.templatetags.unicorn import unicorn
from django_unicorn.utils import generate_checksum
from example.coffee.models import Flavor


class FakeModel:
    # Fake a model that only has an id

    def __init__(self):
        self.id = 178

    def to_json(self):
        return {"id": self.id}


class FakeComponent(UnicornView):
    template_name = "templates/test_component.html"


class FakeComponentParent(UnicornView):
    template_name = "templates/test_component_parent.html"


class FakeComponentParentImplicit(UnicornView):
    template_name = "templates/test_component_parent_implicit.html"


class FakeComponentChild(UnicornView):
    template_name = "templates/test_component_child.html"


class FakeComponentChildImplicit(UnicornView):
    template_name = "templates/test_component_child_implicit.html"

    def has_parent(self):
        return self.parent is not None


class FakeComponentArgs(UnicornView):
    template_name = "templates/test_component_args.html"
    hello = "world"

    def mount(self):
        if self.component_args:
            self.hello = self.component_args[0]


class FakeComponentKwargs(UnicornView):
    template_name = "templates/test_component_kwargs.html"
    hello = "world"

    def mount(self):
        if self.component_kwargs is not None:
            self.hello = self.component_kwargs.get("test_kwarg")


class FakeComponentKwargsWithHtmlEntity(UnicornView):
    template_name = "templates/test_component_kwargs_with_html_entity.html"
    hello = "world"

    def mount(self):
        if self.component_kwargs is not None:
            self.hello = self.component_kwargs.get("test_kwarg")


class FakeComponentModel(UnicornView):
    template_name = "templates/test_component_model.html"
    model_id = None

    def mount(self):
        if self.component_kwargs:
            self.model_id = self.component_kwargs.get("model_id")


class FakeComponentCalls(UnicornView):
    template_name = "templates/test_component.html"

    def mount(self):
        self.call("testCall")


class FakeComponentCalls2(UnicornView):
    template_name = "templates/test_component.html"

    def mount(self):
        self.call("testCall2", "hello")


def test_unicorn_template_renders(client):
    response = client.get("/test")
    content = response.rendered_content.strip()

    assert response.wsgi_request.path == "/test"
    assert "WSGIRequest" in content
    assert content.startswith("<div unicorn:id")
    assert 'unicorn:name="tests.templatetags.test_unicorn_render.FakeComponentKwargs"' in content
    assert '<script type="application/json" id="unicorn:data:' in content


def test_unicorn_template_renders_with_parent_and_child(client):
    response = client.get("/test-parent")
    content = response.content.decode().strip()

    assert response.wsgi_request.path == "/test-parent"
    assert content.startswith("<div unicorn:id")
    assert 'unicorn:name="tests.templatetags.test_unicorn_render.FakeComponentParent"' in content
    assert 'unicorn:name="tests.templatetags.test_unicorn_render.FakeComponentChild"' in content
    assert "--parent--" in content
    assert "==child==" in content
    assert '<script type="application/json" id="unicorn:data:' in content


def test_unicorn_template_renders_with_parent_and_child_with_templateview(client):
    response = client.get("/test-parent-template")
    content = response.content.decode().strip()

    assert response.wsgi_request.path == "/test-parent-template"
    assert content.startswith("<div unicorn:id")
    assert 'unicorn:name="tests.templatetags.test_unicorn_render.FakeComponentParent"' in content
    assert 'unicorn:name="tests.templatetags.test_unicorn_render.FakeComponentChild"' in content
    assert "--parent--" in content
    assert "==child==" in content
    assert '<script type="application/json" id="unicorn:data:' in content


def test_unicorn_template_renders_with_implicit_parent_and_child(client):
    response = client.get("/test-parent-implicit")
    content = response.content.decode().strip()

    assert response.wsgi_request.path == "/test-parent-implicit"
    assert content.startswith("<div unicorn:id")
    assert 'unicorn:name="tests.templatetags.test_unicorn_render.FakeComponentParentImplicit"' in content
    assert 'unicorn:name="tests.templatetags.test_unicorn_render.FakeComponentChildImplicit"' in content
    assert "--parent--" in content
    assert "==child==" in content
    assert "has_parent:True" in content
    assert '<script type="application/json" id="unicorn:data:' in content


def test_unicorn_render_arg():
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentArgs' 'tested!'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    actual = unicorn_node.render(Context(context))

    assert "<b>tested!</b>" in actual


def test_unicorn_render_kwarg():
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' test_kwarg='tested!'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    actual = unicorn_node.render(Context(context))

    assert "<b>tested!</b>" in actual


def test_unicorn_render_invalid_kwarg():
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' test_kwarg=tested",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    actual = unicorn_node.render(Context(context))

    assert "<b>None</b>" in actual


def test_unicorn_render_context_variable():
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' test_kwarg=test_var.nested",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {"test_var": {"nested": "variable!"}}
    actual = unicorn_node.render(Context(context))

    assert "<b>variable!</b>" in actual


def test_unicorn_render_with_invalid_html():
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargsWithHtmlEntity' test_kwarg=test_var.nested",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {"test_var": {"nested": "variable!"}}
    actual = unicorn_node.render(Context(context))

    assert "-&gt;variable!&lt;-" in actual


def test_unicorn_render_parent(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' parent=view",
    )
    unicorn_node = unicorn(Parser([]), token)
    view = FakeComponentParent(component_name="test", component_id="test_unicorn_render_parent")
    context = {"view": view}
    unicorn_node.render(Context(context))

    assert unicorn_node.parent
    assert (
        unicorn_node.component_id
        == "test_unicorn_render_parent:tests.templatetags.test_unicorn_render.FakeComponentKwargs"
    )


def test_unicorn_render_parent_with_key(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' parent=view key='blob'",
    )
    unicorn_node = unicorn(Parser([]), token)
    view = FakeComponentParent(component_name="test", component_id="test_unicorn_render_parent_with_key")
    context = {"view": view}
    unicorn_node.render(Context(context))

    assert (
        unicorn_node.component_id
        == "test_unicorn_render_parent_with_key:tests.templatetags.test_unicorn_render.FakeComponentKwargs:blob"
    )


def test_unicorn_render_parent_with_id(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' parent=view id='flob'",
    )
    unicorn_node = unicorn(Parser([]), token)
    view = FakeComponentParent(component_name="test", component_id="test_unicorn_render_parent_with_id")
    context = {"view": view}
    unicorn_node.render(Context(context))

    assert (
        unicorn_node.component_id
        == "test_unicorn_render_parent_with_id:tests.templatetags.test_unicorn_render.FakeComponentKwargs:flob"
    )


def test_unicorn_render_parent_with_pk(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' parent=view pk=99",
    )
    unicorn_node = unicorn(Parser([]), token)
    view = FakeComponentParent(component_name="test", component_id="test_unicorn_render_parent_with_pk")
    context = {"view": view}
    unicorn_node.render(Context(context))

    assert (
        unicorn_node.component_id
        == "test_unicorn_render_parent_with_pk:tests.templatetags.test_unicorn_render.FakeComponentKwargs:99"
    )


def test_unicorn_render_parent_with_model_id(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' parent=view model=model",
    )
    unicorn_node = unicorn(Parser([]), token)
    view = FakeComponentParent(component_name="test", component_id="test_unicorn_render_parent_with_model_id")

    context = {"view": view, "model": FakeModel()}
    unicorn_node.render(Context(context))

    assert (
        unicorn_node.component_id
        == "test_unicorn_render_parent_with_model_id:tests.templatetags.test_unicorn_render.FakeComponentKwargs:178"
    )


@pytest.mark.django_db
def test_unicorn_render_parent_with_model_pk(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' parent=view model=model",
    )
    unicorn_node = unicorn(Parser([]), token)
    view = FakeComponentParent(component_name="test", component_id="test_unicorn_render_parent_with_model_pk")

    flavor = Flavor(pk=187)

    context = {"view": view, "model": flavor}
    unicorn_node.render(Context(context))

    assert (
        unicorn_node.component_id
        == "test_unicorn_render_parent_with_model_pk:tests.templatetags.test_unicorn_render.FakeComponentKwargs:187"
    )


def test_unicorn_render_id_use_pk():
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentModel' model_id=model.id",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {"model": {"pk": 123}}
    actual = unicorn_node.render(Context(context))

    assert "==123==" in actual


def test_unicorn_render_component_one_script_tag(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1


def test_unicorn_render_component_minify_html(settings):
    settings.DEBUG = True
    settings.UNICORN = {**settings.UNICORN, "MINIFY_HTML": True}
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert "<script type=module" in html
    assert len(re.findall("<script type=module", html)) == 1

    settings.UNICORN = {**settings.UNICORN, "MINIFY_HTML": False}


def test_unicorn_render_child_component_no_script_tag(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentKwargs' parent=view",
    )
    unicorn_node = unicorn(Parser([]), token)
    view = FakeComponentParent(
        component_name="test-no-script", component_id="test_unicorn_render_child_component_no_script_tag"
    )
    context = {"view": view}
    html = unicorn_node.render(Context(context))

    assert "<script" not in html


def test_unicorn_render_parent_component_one_script_tag(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentParent'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1


def test_unicorn_render_calls(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentCalls'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1
    assert '"calls":[{"fn":"testCall","args":[]}]' in html


def test_unicorn_render_calls_with_arg(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentCalls2'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1
    assert '"calls":[{"fn":"testCall2","args":["hello"]}]' in html


def test_unicorn_render_calls_no_mount_call(settings):
    settings.DEBUG = True
    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentParent'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1
    assert '"calls":[]' in html


def test_unicorn_render_hash(settings):
    settings.DEBUG = True
    settings.UNICORN = {**settings.UNICORN, "SCRIPT_LOCATION": "after"}

    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentParent'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1
    assert '"hash":"' in html

    # Assert that the content hash is correct
    script_idx = html.index("<script")
    rendered_content_without_inserted_after_script = html[:script_idx]
    expected_hash = generate_checksum(rendered_content_without_inserted_after_script)
    assert f'"hash":"{expected_hash}"' in html


def test_unicorn_render_hash_append(settings):
    settings.DEBUG = True
    settings.UNICORN = {**settings.UNICORN, "SCRIPT_LOCATION": "append"}

    token = Token(
        TokenType.TEXT,
        "unicorn 'tests.templatetags.test_unicorn_render.FakeComponentParent'",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1
    assert '"hash":"' in html

    # Assert that the content hash is correct
    script_idx = html.index("<script")
    rendered_content_without_appended_script = html[:script_idx] + "</div>"
    expected_hash = generate_checksum(rendered_content_without_appended_script)
    assert f'"hash":"{expected_hash}"' in html


def test_unicorn_render_with_component_name_from_context():
    token = Token(
        TokenType.TEXT,
        "unicorn component_name",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {"component_name": "tests.templatetags.test_unicorn_render.FakeComponent"}
    html = unicorn_node.render(Context(context))

    assert '<script type="module"' in html
    assert len(re.findall('<script type="module"', html)) == 1


def test_unicorn_render_with_invalid_component_name_from_context():
    token = Token(
        TokenType.TEXT,
        "unicorn bad_component_name",
    )
    unicorn_node = unicorn(Parser([]), token)
    context = {"component_name": "tests.templatetags.test_unicorn_render.FakeComponent"}

    with pytest.raises(ComponentNotValidError) as e:
        unicorn_node.render(Context(context))

    assert (
        e.exconly()
        == "django_unicorn.errors.ComponentNotValidError: Component template is not valid: bad_component_name."
    )



import time
from typing import Any

import orjson
import shortuuid
from tests.views.message.utils import post_and_get_response

from django_unicorn.components import UnicornView, unicorn_view
from django_unicorn.utils import generate_checksum


def _post_to_component(
    client,
    method_name: str,
    component_name: str = "FakeComponent",
    data: dict | None = None,
) -> Any:
    if data is None:
        data = {}

    response = post_and_get_response(
        client,
        url=f"/message/tests.views.fake_components.{component_name}",
        data=data,
        action_queue=[
            {
                "payload": {"name": method_name},
                "type": "callMethod",
            }
        ],
    )

    return response


def test_message_call_method(client):
    data = {"method_count": 0}
    body = _post_to_component(client, "test_method", data=data)

    assert body["data"].get("method_count") == 1


def test_message_call_method_with_dictionary_checksum(client):
    data = {"dictionary": {"1": "test", "2": "anothertest", "3": "", "4": "moretest"}}
    body = _post_to_component(client, "test_method", data=data)

    assert not body["errors"]


def test_message_call_method_redirect(client):
    body = _post_to_component(client, "test_redirect")

    assert "redirect" in body
    redirect = body["redirect"]
    assert redirect.get("url") == "/something-here"
    assert redirect.get("refresh", False) is False


def test_message_call_method_with_message(client):
    body = _post_to_component(client, method_name="test_message", component_name="FakeComponentWithMessage")

    assert not body["errors"]

    assert "dom" in body
    dom = body["dom"]

    assert "test success" in dom


def test_message_call_method_redirect_with_message(client):
    data = {}
    message = {
        "actionQueue": [
            {
                "payload": {"name": "test_redirect_with_message"},
                "type": "callMethod",
            }
        ],
        "data": data,
        "checksum": generate_checksum(str(data)),
        "id": shortuuid.uuid()[:8],
        "epoch": time.time(),
    }

    response = client.post(
        "/message/tests.views.fake_components.FakeComponentWithMessage",
        message,
        content_type="application/json",
    )

    body = orjson.loads(response.content)

    assert "redirect" in body
    redirect = body["redirect"]
    assert redirect.get("url") == "/something-here"
    assert redirect.get("refresh", False) is False

    assert "dom" in body
    dom = body["dom"]

    # Check that the message wasn't rendered out because redirects should defer messages until the next render
    assert "test success" not in dom

    # Check the private variable to ensure that there is a queue message for next render
    assert len(response.wsgi_request._messages._queued_messages) == 1


def test_message_call_method_refresh_redirect(client):
    body = _post_to_component(client, method_name="test_refresh_redirect")

    assert "redirect" in body
    redirect = body["redirect"]
    assert redirect.get("url") == "/something-here"
    assert redirect.get("refresh")
    assert redirect.get("title") == "new title"


def test_message_call_method_hash_update(client):
    body = _post_to_component(client, method_name="test_hash_update")

    assert "redirect" in body
    assert body["redirect"].get("hash") == "#test=1"


def test_message_call_method_return_value(client):
    body = _post_to_component(client, method_name="test_return_value")

    assert "return" in body
    return_data = body["return"]
    assert return_data.get("method") == "test_return_value"
    assert return_data.get("args") == []
    assert return_data.get("kwargs") == {}
    assert return_data.get("value") == "booya"


def test_message_call_method_poll_update(client):
    body = _post_to_component(client, method_name="test_poll_update")

    assert "poll" in body
    poll = body["poll"]
    assert poll.get("timing") == 1000
    assert poll.get("disable") is True
    assert poll.get("method") == "new_method"


def test_message_call_method_setter(client):
    data = {"method_count": 0}
    body = _post_to_component(client, method_name="method_count=2", data=data)

    assert body["data"].get("method_count") == 2


def test_message_call_method_nested_setter(client):
    data = {"nested": {"check": True}}
    body = _post_to_component(client, method_name="nested.check=False", data=data)

    assert body["data"].get("nested").get("check") is False


def test_message_call_method_multiple_nested_setter(client):
    data = {"nested": {"another": {"bool": True}}}
    body = _post_to_component(client, method_name="nested.another.bool=False", data=data)

    assert body["data"].get("nested").get("another").get("bool") is False


def test_message_call_method_toggle(client):
    data = {"check": False}
    body = _post_to_component(client, method_name="$toggle('check')", data=data)

    assert body["data"].get("check") is True


def test_message_call_method_nested_toggle(client):
    data = {"nested": {"check": False}}
    body = _post_to_component(client, method_name="$toggle('nested.check')", data=data)

    assert body["data"].get("nested").get("check") is True


def test_message_call_method_args(client):
    data = {"method_count": 0}
    body = _post_to_component(client, method_name="test_method_args(3)", data=data)

    assert body["data"].get("method_count") == 3


def test_message_call_method_kwargs(client):
    data = {"method_count": 0}
    body = _post_to_component(client, method_name="test_method_kwargs(count=99)", data=data)

    assert body["data"].get("method_count") == 99


def test_message_call_method_no_validation(client):
    body = _post_to_component(
        client,
        method_name="set_text_no_validation",
        component_name="FakeValidationComponent",
    )

    assert not body["errors"]


def test_message_call_method_validation(client):
    body = _post_to_component(
        client,
        method_name="set_text_with_validation",
        component_name="FakeValidationComponent",
    )

    assert body["errors"]
    assert body["errors"]["number"]
    assert body["errors"]["number"][0]["code"] == "required"
    assert body["errors"]["number"][0]["message"] == "This field is required."


def test_message_call_method_reset(client):
    data = {"method_count": 1}

    body = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {"payload": {"name": "method_count=2"}, "type": "callMethod"},
            {
                "payload": {"name": "$reset"},
                "type": "callMethod",
            },
        ],
    )

    assert body["data"]["method_count"] == 0
    # `data` should contain all data (not just the diffs) for resets
    assert body["data"].get("check") is not None
    assert body["data"].get("dictionary") is not None


def test_message_call_method_refresh(client):
    data = {"method_count": 1}
    body = _post_to_component(client, method_name="$refresh", data=data)

    assert body["data"]["method_count"] == 1
    # `data` should contain all data (not just the diffs) for refreshes
    assert body["data"].get("check") is not None
    assert body["data"].get("dictionary") is not None


def test_message_call_method_caches_disabled(client, monkeypatch, settings):
    monkeypatch.setattr(unicorn_view, "COMPONENTS_MODULE_CACHE_ENABLED", False)
    settings.CACHES = {
        **settings.CACHES,
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        },
    }

    component_id = shortuuid.uuid()[:8]
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data={"method_count": 0},
        action_queue=[
            {
                "payload": {"name": "test_method"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
    )

    method_count = response["data"].get("method_count")

    assert method_count == 1

    # Get the component again
    view = UnicornView.create(
        component_name="tests.views.fake_components.FakeComponent",
        component_id=component_id,
        use_cache=True,
    )

    # Component is not retrieved from any caches
    assert view.method_count == 0


def test_message_call_method_module_cache_disabled(client, monkeypatch, settings):
    monkeypatch.setattr(unicorn_view, "COMPONENTS_MODULE_CACHE_ENABLED", False)
    settings.UNICORN = {**settings.UNICORN, "CACHE_ALIAS": "default"}
    settings.CACHES = {
        **settings.CACHES,
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        },
    }

    component_id = shortuuid.uuid()[:8]
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data={"method_count": 0},
        action_queue=[
            {
                "payload": {"name": "test_method"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
    )

    method_count = response["data"].get("method_count")

    assert method_count == 1

    # Get the component again and it should be found in local memory cache
    view = UnicornView.create(
        component_name="tests.views.fake_components.FakeComponent",
        component_id=component_id,
        use_cache=True,
    )

    # Component is retrieved from the local memory cache
    assert view.method_count == 0


def test_message_call_method_cache_backend_dummy(client, monkeypatch, settings):
    monkeypatch.setattr(unicorn_view, "COMPONENTS_MODULE_CACHE_ENABLED", True)
    settings.CACHES = {
        **settings.CACHES,
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        },
    }

    component_id = shortuuid.uuid()[:8]
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data={"method_count": 0},
        action_queue=[
            {
                "payload": {"name": "test_method"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
    )

    method_count = response["data"].get("method_count")

    assert method_count == 1

    # Get the component again and it should be found in local memory cache
    view = UnicornView.create(
        component_name="tests.views.fake_components.FakeComponent",
        component_id=component_id,
        use_cache=True,
    )

    # Component is retrieved from the module cache
    assert view.method_count == method_count


def test_message_call_method_validation_error(client):
    body = _post_to_component(client, "test_validation_error")

    assert body["errors"]
    assert body["errors"]["check"]
    assert body["errors"]["check"][0]["code"] == "required"
    assert body["errors"]["check"][0]["message"]
    assert body["errors"]["check"][0]["message"] == ["Check is required"]


def test_message_call_method_validation_error_list(client):
    body = _post_to_component(client, "test_validation_error_list")

    assert body["errors"]
    assert body["errors"]["__all__"] == [{"code": "required", "message": "Check is required"}]


def test_message_call_method_validation_error_list_no_code(client):
    body = _post_to_component(client, "test_validation_error_list_no_code")

    assert body["error"]
    assert body["error"] == "Error code must be specified"


def test_message_call_method_validation_error_no_code(client):
    body = _post_to_component(client, "test_validation_error_no_code")

    assert body["error"]
    assert body["error"] == "Error code must be specified"


def test_message_call_method_validation_error_string(client):
    body = _post_to_component(client, "test_validation_error_string")

    assert body["errors"]
    assert body["errors"]["__all__"] == [{"code": "required", "message": "Check is required"}]


def test_message_call_method_validation_error_string_no_code(client):
    body = _post_to_component(client, "test_validation_error_string_no_code")

    assert body["error"]
    assert body["error"] == "Error code must be specified"



# Generated by Django 3.0.8 on 2020-09-29 02:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Flavor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('label', models.CharField(max_length=255)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='coffee.Flavor')),
            ],
        ),
    ]



# Generated by Django 4.2.5 on 2023-10-29 22:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("books", "0003_auto_20221110_0400"),
    ]

    operations = [
        migrations.AddField(
            model_name="book",
            name="type",
            field=models.IntegerField(choices=[(1, "Hardcover"), (2, "Softcover")], default=1),
        ),
    ]



from django.template.base import Parser, Token, TokenType

from django_unicorn.templatetags.unicorn import unicorn


def test_unicorn():
    token = Token(TokenType.TEXT, "unicorn 'todo'")
    unicorn_node = unicorn(Parser([]), token)

    assert unicorn_node.component_name.resolve({}) == "todo"
    assert len(unicorn_node.kwargs) == 0


def test_unicorn_kwargs():
    token = Token(TokenType.TEXT, "unicorn 'todo' blob='blob'")
    unicorn_node = unicorn(Parser([]), token)

    assert unicorn_node.kwargs == {"blob": "blob"}


def test_unicorn_args_and_kwargs():
    # args after the component name get ignored
    token = Token(TokenType.TEXT, "unicorn 'todo' '1' 2 hello='world' test=3 '4'")
    unicorn_node = unicorn(Parser([]), token)

    assert unicorn_node.kwargs == {"hello": "world", "test": 3}



from django.utils.timezone import now

from django_unicorn.components import UnicornView


class JsView(UnicornView):
    states = (
        "Alabama",
        "Alaska",
        "Wisconsin",
        "Wyoming",
    )
    selected_state = ""
    select2_datetime = now()
    scroll_counter = 0
    load_js = False

    def call_javascript(self):
        self.call("callAlert", "world")

    def call_javascript_module(self):
        self.call("HelloJs.hello", "world!")

    def get_now(self):
        self.select2_datetime = now()

    def change_states(self):
        self.states = ("Pennsylvania",)

    def select_state(self, val, idx):
        print("select_state called", val)
        print("select_state called idx", idx)
        self.selected_state = val

    def increase_counter(self):
        if self.scroll_counter >= 2:
            return False

        self.scroll_counter += 1

    class Meta:
        javascript_excludes = ("states",)



import pytest
from django.conf import settings


def pytest_configure():
    templates = [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": ["tests"],
            "OPTIONS": {
                "context_processors": [
                    "django.template.context_processors.request",
                    "django.contrib.messages.context_processors.messages",
                ],
            },
        }
    ]

    databases = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
        }
    }

    installed_apps = [
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.contenttypes",
        "django.contrib.auth",
        "django_unicorn",
        "example.coffee.apps.Config",
        "example.books.apps.Config",
    ]

    unicorn_settings = {
        "SERIAL": {"ENABLED": True, "TIMEOUT": 5},
        "CACHE_ALIAS": "default",
        "APPS": ("unicorn",),
        "MINIFY_HTML": False,
        "SCRIPT_LOCATION": "after",
    }

    caches = {
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
            "LOCATION": "unique-snowflake",
        }
    }

    settings.configure(
        DEBUG=True,
        SECRET_KEY="this-is-a-secret",
        TEMPLATES=templates,
        MIDDLEWARE=[
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.contrib.messages.middleware.MessageMiddleware",
        ],
        ROOT_URLCONF="tests.urls",
        DATABASES=databases,
        INSTALLED_APPS=installed_apps,
        UNIT_TEST=True,
        UNICORN=unicorn_settings,
        CACHES=caches,
        MESSAGE_STORAGE="django.contrib.messages.storage.session.SessionStorage",
        SESSION_ENGINE="django.contrib.sessions.backends.file",
    )


@pytest.fixture(autouse=True)
def reset_settings(settings):
    """
    This takes the original `UNICORN` settings before the test is run, runs the test, and then resets them afterwards.
    This is required because mutating nested dictionaries does not reset them as expected by `pytest-django`.
    More details in https://github.com/pytest-dev/pytest-django/issues/601#issuecomment-440676001.
    """

    # Get original settings
    cache_settings = {**settings.CACHES}
    unicorn_settings = {**settings.UNICORN}
    django_unicorn_settings = {}

    if hasattr(settings, "DJANGO_UNICORN"):
        django_unicorn_settings = {**settings.DJANGO_UNICORN}

    # Run test
    yield

    # Re-set original settings
    settings.CACHES = cache_settings
    settings.UNICORN = unicorn_settings

    if django_unicorn_settings:
        settings.DJANGO_UNICORN = django_unicorn_settings



from tests.views.message.utils import post_and_get_response


def test_message_nested_sync_input(client):
    data = {"dictionary": {"name": "test"}}
    action_queue = [
        {
            "payload": {"name": "dictionary.name", "value": "test1"},
            "type": "syncInput",
        }
    ]
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=action_queue,
    )

    assert not response["errors"]
    assert response["data"].get("dictionary") == {"name": "test1"}


def test_message_sync_input_choices_with_select_widget(client):
    """
    ModelForms with a Model that have a field with `choices` and the form's field uses a Select widget.
    Need to handle Select widget specifically otherwise `field.widget.format_value` will return a list
    that only contains one object.
    """

    data = {"type": 1}
    action_queue = [
        {
            "payload": {"name": "type", "value": 2},
            "type": "syncInput",
        }
    ]
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeModelFormComponent",
        data=data,
        action_queue=action_queue,
    )

    assert not response["errors"]
    assert response["data"].get("type") == 2



from tests.views.message.utils import post_and_get_response

from django_unicorn.components import UnicornView


class FakeComponent(UnicornView):
    template_name = "templates/test_component_variable.html"

    hello = ""


class FakeComponentSafe(UnicornView):
    template_name = "templates/test_component_variable.html"

    hello = ""

    class Meta:
        safe = ("hello",)


def test_html_entities_encoded(client):
    data = {"hello": "test"}
    action_queue = [
        {
            "payload": {"name": "hello", "value": "<b>test1</b>"},
            "type": "syncInput",
        }
    ]
    response = post_and_get_response(
        client,
        url="/message/tests.views.test_process_component_request.FakeComponent",
        data=data,
        action_queue=action_queue,
    )

    assert not response["errors"]
    assert response["data"].get("hello") == "<b>test1</b>"
    assert "&lt;b&gt;test1&lt;/b&gt;" in response["dom"]


def test_safe_html_entities_not_encoded(client):
    data = {"hello": "test"}
    action_queue = [
        {
            "payload": {"name": "hello", "value": "<b>test1</b>"},
            "type": "syncInput",
        }
    ]
    response = post_and_get_response(
        client,
        url="/message/tests.views.test_process_component_request.FakeComponentSafe",
        data=data,
        action_queue=action_queue,
    )

    assert not response["errors"]
    assert response["data"].get("hello") == "<b>test1</b>"
    assert "<b>test1</b>" in response["dom"]



from unittest.mock import MagicMock, patch

from django_unicorn.cacher import (
    CacheableComponent,
    cache_full_tree,
    restore_from_cache,
)
from django_unicorn.components import UnicornView


class FakeComponent(UnicornView):
    pass


class FakeComponentWithTemplateHtml(UnicornView):
    template_html = """<div>
    testing
</div>
"""


def test_cacheable_component_request_is_none_then_restored():
    component = FakeComponent(
        component_id="test_cacheable_component_request_is_none_then_restored", component_name="hello-world"
    )
    request = component.request = MagicMock()
    assert component.request

    with CacheableComponent(component):
        assert component.request is None

    assert component.request == request


def test_cacheable_component_extra_context_is_none_then_restored():
    component = FakeComponent(
        component_id="test_cacheable_component_extra_context_is_none_then_restored", component_name="hello-world"
    )
    extra_context = component.extra_context = MagicMock()
    assert component.extra_context

    with CacheableComponent(component):
        assert component.extra_context is None

    assert component.extra_context == extra_context


def test_cacheable_component_parents_have_request_restored():
    component = FakeComponent(
        component_id="test_cacheable_component_parents_have_request_restored_1", component_name="hello-world"
    )
    component2 = FakeComponent(
        component_id="test_cacheable_component_parents_have_request_restored_2",
        component_name="hello-world",
        parent=component,
    )
    component3 = FakeComponent(
        component_id="test_cacheable_component_parents_have_request_restored_3",
        component_name="hello-world",
        parent=component2,
    )
    request = MagicMock()
    extra_content = "extra_content"
    for c in [component, component2, component3]:
        c.request = request
        c.extra_context = extra_content

    with CacheableComponent(component3):
        assert component.request is None
        assert component2.request is None
        assert component3.request is None
        assert component.extra_context is None
        assert component2.extra_context is None
        assert component3.extra_context is None

    assert component.request == request
    assert component2.request == request
    assert component3.request == request
    assert component.extra_context == extra_content
    assert component2.extra_context == extra_content
    assert component3.extra_context == extra_content


def test_restore_cached_component_children_have_request_set():
    component = FakeComponent(
        component_id="test_restore_cached_component_children_have_request_set_1", component_name="hello-world"
    )
    component2 = FakeComponent(
        component_id="test_restore_cached_component_children_have_request_set_2", component_name="hello-world"
    )
    component3 = FakeComponent(
        component_id="test_restore_cached_component_children_have_request_set_3", component_name="hello-world"
    )
    component4 = FakeComponent(
        component_id="test_restore_cached_component_children_have_request_set_4", component_name="hello-world"
    )
    component3.children.append(component4)
    component.children.extend([component2, component3])
    request = MagicMock()
    extra_content = "extra_content"
    for c in [component, component2, component3, component4]:
        c.request = request
        c.extra_context = extra_content

    with CacheableComponent(component):
        assert component.request is None
        assert component2.request is None
        assert component3.request is None
        assert component4.request is None
        assert component.extra_context is None
        assert component2.extra_context is None
        assert component3.extra_context is None
        assert component4.extra_context is None

    assert component.request == request
    assert component2.request == request
    assert component3.request == request
    assert component4.request == request
    assert component.extra_context == extra_content
    assert component2.extra_context == extra_content
    assert component3.extra_context == extra_content
    assert component4.extra_context == extra_content


class ExampleCachingComponent(UnicornView):
    name = "World"

    def get_name(self):
        return "World"

    def __str__(self):
        return f"ECC {self.component_id} - {self.component_name}"


def test_caching_components(settings):
    settings.CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
            "LOCATION": "unique-snowflake",
        }
    }
    settings.UNICORN = {**settings.UNICORN, "CACHE_ALIAS": "default"}
    root = ExampleCachingComponent(component_id="test_caching_components_1", component_name="root")
    child1 = ExampleCachingComponent(
        component_id="test_caching_components_child_1", component_name="child1", parent=root
    )
    child2 = ExampleCachingComponent(
        component_id="test_caching_components_child_2", component_name="child2", parent=root
    )
    child3 = ExampleCachingComponent(
        component_id="test_caching_components_child_3", component_name="child3", parent=root
    )
    grandchild = ExampleCachingComponent(
        component_id="test_caching_components_grandchild_1", component_name="grandchild", parent=child1
    )
    grandchild2 = ExampleCachingComponent(
        component_id="test_caching_components_grandchild_2", component_name="grandchild2", parent=child1
    )
    grandchild3 = ExampleCachingComponent(
        component_id="test_caching_components_grandchild_3", component_name="grandchild3", parent=child3
    )

    cache_full_tree(child3)
    request = MagicMock()

    restored: UnicornView = restore_from_cache(child2.component_cache_key, request)

    restored_root = restored
    while restored_root.parent:
        restored_root = restored_root.parent

    assert root.component_id == restored_root.component_id
    assert 3 == len(restored_root.children)

    for i, child in enumerate([child1, child2, child3]):
        assert restored_root.children[i].component_id == child.component_id

    assert 2 == len(restored_root.children[0].children)

    for i, child in enumerate([grandchild, grandchild2]):
        assert restored_root.children[0].children[i].component_id == child.component_id

    assert not restored_root.children[1].children
    assert 1 == len(restored_root.children[2].children)
    assert grandchild3.component_id == restored_root.children[2].children[0].component_id


@patch("django_unicorn.cacher.create_template")
def test_caching_components_with_template_html(create_template):
    component = FakeComponentWithTemplateHtml(
        component_id="test_caching_components_with_template_html", component_name="template-html-test"
    )
    assert component.template_html

    # manually set template_name to None which happens outside of this code normally
    component.template_name = None

    # Caching will pop the `Template` instance from component.template_name when it enters,
    # and re-create it when it exits
    cache_full_tree(component)

    assert component.template_html
    assert isinstance(component.template_html, str)

    # check that template_name will re-create a Template based on `template_html`
    create_template.assert_called_once_with(component.template_html)



import toml  # type: ignore

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys


# sys.path.insert(0, os.path.abspath("....django-unicorn"))


# -- Project information -----------------------------------------------------

project = "Unicorn"
copyright = "2026, Adam Hill"  # noqa: A001
author = "Adam Hill"

pyproject = toml.load("../../pyproject.toml")
version = pyproject["project"]["version"]
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "myst_parser",
    "sphinx_design",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "rst2pdf.pdfbuilder",
    "autoapi.extension",
    "sphinxext.opengraph",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 3

myst_enable_extensions = ["colon_fence"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []
html_title = "Unicorn"

myst_heading_anchors = 3
myst_enable_extensions = ["linkify", "colon_fence"]

pdf_documents = [
    ("index", "unicorn-latest", "Unicorn", "Adam Hill"),
]

autoapi_dirs = [
    "../../django_unicorn",
]
autoapi_root = "api"
autoapi_add_toctree_entry = False
autoapi_generate_api_docs = True
# autoapi_keep_files = True  # useful for debugging generated errors
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
]
autoapi_type = "python"
autodoc_typehints = "signature"


def skip_member(app, what, name, obj, skip, options):  # noqa: ARG001
    if what == "data" and name.endswith(".logger"):
        skip = True
    elif "startunicorn" in name:
        skip = True

    return skip


def setup(sphinx):
    sphinx.connect("autoapi-skip-member", skip_member)


ogp_site_url = "https://www.django-unicorn.com/"



import pytest

from django_unicorn.typer import _construct_model
from example.books.models import Author, Book
from example.coffee.models import Flavor


def test_construct_model_simple_model():
    model_data = {
        "pk": 1,
        "name": "test-name",
    }

    actual = _construct_model(Flavor, model_data)

    assert actual.pk == 1
    assert actual.name == "test-name"


@pytest.mark.django_db
def test_construct_model_foreign_key():
    flavor = Flavor(name="first-flavor")
    flavor.save()
    parent = Flavor(name="parent-flavor")
    parent.save()

    model_data = {
        "pk": flavor.pk,
        "name": flavor.name,
        "parent": parent.pk,
    }

    actual = _construct_model(Flavor, model_data)

    assert actual.pk == flavor.pk
    assert actual.name == flavor.name
    assert actual.parent.pk == parent.pk
    assert actual.parent.name == parent.name


@pytest.mark.django_db
@pytest.mark.skip("This test isn't all that helpful unless related models get serialized")
def test_construct_model_recursive_foreign_key():
    flavor = Flavor(name="first-flavor")
    flavor.save()
    parent = Flavor(name="parent-flavor")
    parent.save()

    model_data = {
        "pk": flavor.pk,
        "name": flavor.name,
        "parent": parent.pk,
    }

    actual = _construct_model(Flavor, model_data)

    assert actual.pk == flavor.pk
    assert actual.name == flavor.name
    assert actual.parent.pk == parent.pk
    assert actual.parent.name == parent.name


@pytest.mark.django_db
def test_construct_model_many_to_many():
    author = Author(name="author 1")
    author.save()
    book = Book(title="book 1", date_published="2021-01-01")
    book.save()
    author.books.add(book)

    author_data = {
        "pk": author.pk,
        "name": author.name,
        "books": [
            book.pk,
        ],
    }

    actual = _construct_model(Author, author_data)

    assert actual.pk == author.pk
    assert actual.name == author.name
    assert actual.books.count() == 1
    assert actual.books.all()[0].title == book.title



from dataclasses import dataclass

import pytest

from django_unicorn.components import UnicornView
from django_unicorn.views.action_parsers.utils import set_property_value
from example.coffee.models import Flavor


@dataclass
class InventoryItem:
    """Class for keeping track of an item in inventory."""

    name: str
    unit_price: float
    quantity_on_hand: int = 0


class PropertyView(UnicornView):
    inventory = InventoryItem("Hammer", 20)


class FakeComponent(UnicornView):
    flavors = []  # noqa: RUF012

    def __init__(self, **kwargs):
        self.flavors = list(Flavor.objects.all())

        super().__init__(**kwargs)


def test_set_property_value_dataclass():
    component = PropertyView(component_name="test", component_id="test_set_property_value_dataclass")
    assert InventoryItem("Hammer", 20) == component.inventory

    set_property_value(
        component,
        "inventory",
        InventoryItem("Hammer", 20, 1),
        {"inventory": InventoryItem("Hammer", 20, 1)},
    )

    assert InventoryItem("Hammer", 20, 1) == component.inventory


@pytest.mark.django_db
def test_set_property_value_array():
    flavor_one = Flavor(name="initial 1")
    flavor_one.save()
    flavor_two = Flavor(name="initial 2")
    flavor_two.save()
    component = FakeComponent(component_name="test", component_id="test_set_property_value_array")

    set_property_value(
        component,
        "flavors.0.name",
        "test 1",
        {"flavors": [{"name": "test"}, {"name": "something"}]},
    )

    assert component.flavors[0].name == "test 1"


@pytest.mark.django_db
def test_set_property_value_foreign_key():
    flavor = Flavor(name="initial 1")
    flavor.save()
    parent = Flavor(name="initial 2")
    parent.save()
    component = FakeComponent(component_name="test", component_id="test_set_property_value_foreign_key")

    set_property_value(
        component,
        "flavors.0.parent",
        parent.pk,
        {
            "flavors": [
                {"name": flavor.name, "parent": None},
            ]
        },
    )

    assert component.flavors[0].parent_id == parent.pk

    # This fails for Django 2.2, but not 3.0
    # assert component.flavors[0].parent.pk == parent.pk



from tests.views.message.utils import post_and_get_response

from django_unicorn.components import UnicornView


class FakeCallsComponent(UnicornView):
    template_name = "templates/test_component.html"

    def test_call(self):
        self.call("testCall")

    def test_call2(self):
        self.call("testCall2")

    def test_call3(self):
        self.call("testCall3", "hello")


FAKE_CALLS_COMPONENT_URL = "/message/tests.views.message.test_calls.FakeCallsComponent"


def test_message_calls(client):
    action_queue = [
        {
            "payload": {"name": "test_call"},
            "type": "callMethod",
            "target": None,
        }
    ]

    response = post_and_get_response(client, url=FAKE_CALLS_COMPONENT_URL, action_queue=action_queue)

    assert response.get("calls") == [{"args": [], "fn": "testCall"}]


def test_message_multiple_calls(client):
    action_queue = [
        {
            "payload": {"name": "test_call"},
            "type": "callMethod",
            "target": None,
        },
        {
            "payload": {"name": "test_call2"},
            "type": "callMethod",
            "target": None,
        },
    ]
    response = post_and_get_response(client, url=FAKE_CALLS_COMPONENT_URL, action_queue=action_queue)

    assert response.get("calls") == [
        {"args": [], "fn": "testCall"},
        {"args": [], "fn": "testCall2"},
    ]


def test_message_calls_with_arg(client):
    action_queue = [
        {
            "payload": {"name": "test_call3"},
            "type": "callMethod",
            "target": None,
        }
    ]

    response = post_and_get_response(client, url=FAKE_CALLS_COMPONENT_URL, action_queue=action_queue)

    assert response.get("calls") == [{"args": ["hello"], "fn": "testCall3"}]



from datetime import datetime, timezone

from django import forms
from django.contrib import messages
from django.contrib.auth.forms import AuthenticationForm
from django.forms import ValidationError
from django.shortcuts import redirect

from django_unicorn.components import (
    HashUpdate,
    LocationUpdate,
    PollUpdate,
    UnicornView,
)
from example.books.models import Book
from example.coffee.models import Flavor


class FakeComponent(UnicornView):
    template_name = "templates/test_component.html"
    dictionary = {"name": "test"}  # noqa: RUF012
    method_count = 0
    check = False
    nested = {"check": False, "another": {"bool": False}}  # noqa: RUF012
    method_arg = ""

    def test_method(self):
        self.method_count += 1

    def test_method_args(self, count):
        self.method_count = count

    def test_method_string_arg(self, param):
        self.method_count += 1
        self.method_arg = param

    def test_method_kwargs(self, count=-1):
        self.method_count = count

    def test_redirect(self):
        return redirect("/something-here")

    def test_refresh_redirect(self):
        return LocationUpdate(redirect("/something-here"), title="new title")

    def test_hash_update(self):
        return HashUpdate("#test=1")

    def test_return_value(self):
        return "booya"

    def test_poll_update(self):
        return PollUpdate(timing=1000, disable=True, method="new_method")

    def test_validation_error(self):
        raise ValidationError({"check": "Check is required"}, code="required")

    def test_validation_error_no_code(self):
        raise ValidationError({"check": "Check is required"})

    def test_validation_error_string(self):
        raise ValidationError("Check is required", code="required")

    def test_validation_error_string_no_code(self):
        raise ValidationError("Check is required")

    def test_validation_error_list(self):
        raise ValidationError([ValidationError({"check": "Check is required"})], code="required")

    def test_validation_error_list_no_code(self):
        raise ValidationError([ValidationError({"check": "Check is required"})])


class FakeModelForm(forms.ModelForm):
    class Meta:
        model = Book
        fields = ("title", "date_published", "type")


class FakeModelFormComponent(UnicornView):
    template_name = "templates/test_component.html"
    form_class = FakeModelForm

    title = None
    date_published = None
    type = None


class FakeModelComponent(UnicornView):
    template_name = "templates/test_component.html"
    flavors = Flavor.objects.all()

    def hydrate(self):
        self.flavors = Flavor.objects.all()


class FakeValidationForm(forms.Form):
    text = forms.CharField(min_length=3, max_length=10)
    date_time = forms.DateTimeField()
    number = forms.IntegerField()
    permanent = forms.BooleanField()


class FakeValidationComponent(UnicornView):
    template_name = "templates/test_component.html"
    form_class = FakeValidationForm

    text = "hello"
    number = ""
    date_time = datetime(2020, 9, 13, 17, 45, 14, tzinfo=timezone.utc)
    permanent = True

    def set_text_no_validation(self):
        self.text = "no validation"

    def set_text_with_validation(self):
        self.text = "validation 33"
        self.validate()

    def set_number(self, number):
        self.number = number


class FakeAuthenticationComponent(UnicornView):
    template_name = "templates/test_component.html"
    form_class = AuthenticationForm

    username = ""
    password = ""


class FakeComponentWithDictionary(UnicornView):
    template_name = "templates/test_component.html"
    dictionary: dict | None = None

    def test_method(self):
        pass


class FakeComponentWithMessage(UnicornView):
    template_name = "templates/test_component_with_message.html"

    def test_message(self):
        assert self.request, "Expect a request in action methods"
        messages.success(self.request, "test success")

    def test_redirect_with_message(self):
        assert self.request, "Expect a request in action methods"
        messages.success(self.request, "test success")
        return redirect("/something-here")


class FakeComponentWithError(UnicornView):
    template_name = "templates/test_component.html"

    def mount(self):
        print(self.not_a_valid_attribute)  # type: ignore # noqa: T201


count_updating = 0
count_updated = 0


class FakeComponentWithUpdateMethods(UnicornView):
    template_name = "templates/test_component.html"

    count = 0

    def updating_count(self, _):
        global count_updating  # noqa: PLW0603
        count_updating += 1

        if count_updating >= 2:
            raise Exception("updating_count called more than once")

        assert count_updating == 1

    def updated_count(self, _):
        global count_updated  # noqa: PLW0603
        count_updated += 1

        if count_updated >= 2:
            raise Exception("count_updated called more than once")

        assert count_updated == 1


count_resolved = 0


class FakeComponentWithResolveMethods(UnicornView):
    template_name = "templates/test_component.html"

    count = 0

    def resolved_count(self, _):
        global count_resolved  # noqa: PLW0603
        count_resolved += 1

        assert count_resolved == 1, "count_resolved called more than once"



import pytest
from django.db import models

from django_unicorn.components import ModelValueMixin
from django_unicorn.serializer import model_value
from example.coffee.models import Flavor


def test_model_value_all_fields():
    flavor = Flavor(name="flavor-1")

    expected = {
        "date": None,
        "datetime": None,
        "decimal_value": None,
        "duration": None,
        "float_value": None,
        "name": flavor.name,
        "label": "",
        "parent": None,
        "pk": None,
        "time": None,
        "uuid": str(flavor.uuid),
        "taste_set": [],
        "origins": [],
    }

    actual = model_value(flavor)

    assert expected == actual


def test_model_value_one_field():
    expected = {"name": "flavor-1"}

    flavor = Flavor(name="flavor-1")
    actual = model_value(flavor, "name")

    assert expected == actual


@pytest.mark.django_db
def test_model_value_multiple_field():
    expected = {
        "pk": 77,
        "name": "flavor-1",
    }

    flavor = Flavor(name="flavor-1", id=77)
    actual = model_value(flavor, "pk", "name")

    assert expected == actual


class FakeModel(ModelValueMixin, models.Model):
    name = models.CharField(max_length=255)

    class Meta:
        app_label = "tests"


def test_model_value_mixin():
    test_model = FakeModel(name="test-model")
    expected = {"name": test_model.name}

    actual = test_model.value("name")
    assert expected == actual

    actual = model_value(test_model, "name")
    assert expected == actual



# Generated by Django 3.1.3 on 2020-12-05 14:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('coffee', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='flavor',
            name='decimal_value',
            field=models.DecimalField(decimal_places=2, max_digits=10, null=True),
        ),
        migrations.AddField(
            model_name='flavor',
            name='float_value',
            field=models.FloatField(null=True),
        ),
    ]



import time

from django.contrib import messages
from django.utils.timezone import now

from django_unicorn.components import PollUpdate, UnicornView


class PollingView(UnicornView):
    polling_disabled = False
    date_example = now()
    current_time = now()
    counter = 0

    def slow_update(self):
        self.counter += 1
        time.sleep(0.8)  # Simulate slow request

    def get_date(self):
        self.current_time = now()
        self.date_example = now()

        messages.error(self.request, "get_date called :(")

        return PollUpdate(timing=2000, disable=False, method="get_date_2")

    def get_date_2(self):
        self.current_time = now()
        self.date_example = now()

        messages.success(self.request, "get_date2 called :)")

        return PollUpdate(timing=1000, disable=False, method="get_date")



from django.urls import path, re_path

from django_unicorn import views

app_name = "django_unicorn"


urlpatterns = (
    re_path(r"message/(?P<component_name>[\w/\.-]+)", views.message, name="message"),
    path("message", views.message, name="message"),  # Only here to build the correct url in scripts.html
)



# -*- coding: utf-8 -*-

from example.books.models import Book, Author
from django_unicorn.components import UnicornView
from typing import Any, cast

class TableView(UnicornView):
    books = Book.objects.none()

    def mount(self):
        self.load_table()

    def load_table(self):
        print(Book.objects.all())
        if Book.objects.count() == 0:
            for author, title in [
                ("Dr. Seuss", "The Cat in the Hat"),
                ("J.K. Rowling", "Harry Potter and the Philosopher's Stone"),
                ("Roald Dahl", "Matilda"),
                ("Maurice Sendak", "Where the Wild Things Are"),
                ("A.A. Milne", "Winnie the Pooh")
            ]:
                author = Author(name=author).save()
                Book(author=author, title=title).save()

        self.books = Book.objects.all()

        for child in self.children:
            if hasattr(child, "is_editing"):
                cast(Any, child).is_editing = False



import pytest

from django_unicorn.components.unicorn_view import get_locations


@pytest.fixture
def cache_clear():
    get_locations.cache_clear()


@pytest.fixture
def clear_apps(settings):
    unicorn_apps = settings.UNICORN["APPS"]
    unicorn_apps = settings.UNICORN.get("APPS")
    if "APPS" in settings.UNICORN:
        settings.UNICORN = {**settings.UNICORN}
        del settings.UNICORN["APPS"]

    yield

    if unicorn_apps:
        settings.UNICORN = {**settings.UNICORN, "APPS": unicorn_apps}


def test_get_locations_kebab_case(cache_clear):  # noqa: ARG001
    expected = [("unicorn.components.hello_world", "HelloWorldView"), ("components.hello_world", "HelloWorldView")]
    actual = get_locations("hello-world")

    assert expected == actual


def test_get_locations_with_slashes(cache_clear):  # noqa: ARG001
    expected = [("unicorn.components.nested.table", "TableView"), ("components.nested.table", "TableView")]
    actual = get_locations("nested/table")

    assert expected == actual


def test_get_locations_with_dots(cache_clear):  # noqa: ARG001
    expected = [
        ("nested", "table"),
        ("unicorn.components.nested.table", "TableView"),
        ("components.nested.table", "TableView"),
    ]
    actual = get_locations("nested.table")

    assert expected == actual


def test_get_locations_fully_qualified_with_dots(cache_clear):  # noqa: ARG001
    expected = [
        ("project.components.hello_world", "HelloWorldView"),
    ]
    actual = get_locations("project.components.hello_world.HelloWorldView")

    assert expected == actual


def test_get_locations_fully_qualified_with_slashes(cache_clear):  # noqa: ARG001
    expected = [("project.components.hello_world", "HelloWorldView")]
    actual = get_locations("project/components/hello_world.HelloWorldView")

    assert expected == actual


def test_get_locations_fully_qualified_with_dots_ends_in_component(cache_clear):  # noqa: ARG001
    expected = [
        ("project.components.hello_world", "HelloWorldComponent"),
    ]
    actual = get_locations("project.components.hello_world.HelloWorldComponent")

    assert expected == actual


def test_get_locations_fully_qualified_with_dots_does_not_end_in_view(cache_clear):  # noqa: ARG001
    """
    The last 2 entries in here are a mess. This documents how things work correctly even though it is not the ideal.
    """

    expected = [
        ("project.components.hello_world", "HelloWorldThing"),
        (
            "unicorn.components.project.components.hello_world.HelloWorldThing",
            "HelloworldthingView",
        ),
        ("components.project.components.hello_world.HelloWorldThing", "HelloworldthingView"),
    ]
    actual = get_locations("project.components.hello_world.HelloWorldThing")

    assert expected == actual


def test_get_locations_apps_setting_tuple(settings, cache_clear):  # noqa: ARG001
    settings.UNICORN = {**settings.UNICORN, "APPS": ("project",)}

    expected = [
        ("project.components.hello_world", "HelloWorldView"),
        ("components.hello_world", "HelloWorldView"),
    ]
    actual = get_locations("hello-world")

    assert expected == actual


def test_get_locations_apps_setting_list(settings, cache_clear):  # noqa: ARG001
    settings.UNICORN = {
        **settings.UNICORN,
        "APPS": [
            "project",
        ],
    }

    expected = [("project.components.hello_world", "HelloWorldView"), ("components.hello_world", "HelloWorldView")]
    actual = get_locations("hello-world")

    assert expected == actual


def test_get_locations_apps_setting_set(settings, cache_clear):  # noqa: ARG001
    settings.UNICORN = {
        **settings.UNICORN,
        "APPS": {"project"},
    }

    expected = [
        ("project.components.hello_world", "HelloWorldView"),
        ("components.hello_world", "HelloWorldView"),
    ]
    actual = get_locations("hello-world")

    assert expected == actual


def test_get_locations_apps_setting_invalid(settings, cache_clear):  # noqa: ARG001
    settings.UNICORN = {**settings.UNICORN, "APPS": "project"}

    with pytest.raises(AssertionError) as e:
        get_locations("hello-world")

    assert e.type is AssertionError
    settings.UNICORN = {**settings.UNICORN, "APPS": ("unicorn",)}


def test_get_locations_installed_app_with_app_config(settings, clear_apps, cache_clear):  # noqa: ARG001
    settings.INSTALLED_APPS = [
        "example.coffee.apps.Config",
    ]

    expected = [
        ("example.coffee.components.hello_world", "HelloWorldView"),
        ("components.hello_world", "HelloWorldView"),
    ]
    actual = get_locations("hello-world")

    assert expected == actual


def test_get_locations_installed_app_with_apps(settings, clear_apps, cache_clear):  # noqa: ARG001
    # test when the app is in a subdirectory "apps"
    settings.INSTALLED_APPS = [
        "example.apps.main",
    ]
    expected_location = [
        ("example.apps.main.components.sidebar_menu", "SidebarMenuView"),
        ("components.sidebar_menu", "SidebarMenuView"),
    ]
    actual_location = get_locations("sidebar-menu")
    assert expected_location == actual_location



from django_unicorn.components import UnicornView
from django_unicorn.views.action_parsers.utils import set_property_value
from django_unicorn.views.objects import ComponentRequest


def handle(component_request: ComponentRequest, component: UnicornView, payload: dict):
    property_name = payload.get("name")
    property_value = payload.get("value")

    call_resolved_method = True

    # If there is more than one action then only call the resolved methods for the last action in the queue
    if len(component_request.action_queue) > 1:
        call_resolved_method = False
        last_action = component_request.action_queue[-1:][0]

        if last_action.payload.get("name") == property_name and last_action.payload.get("value") == property_value:
            call_resolved_method = True

    set_property_value(
        component, property_name, property_value, component_request.data, call_resolved_method=call_resolved_method
    )



from typing import Optional

from django_unicorn.components import UnicornView
from example.coffee.models import Flavor


class HtmlInputsView(UnicornView):
    is_checked = False
    another_check = True
    thing = "🐙"
    flavor: Optional[Flavor] = None
    flavors = Flavor.objects.none()
    things = [
        "alien",
    ]
    pie = "cherry"
    paragraph = ""
    state = ""

    ALL_STATES = (
        "Alabama",
        "Alaska",
        "Arizona",
        "Arkansas",
        "California",
        "Colorado",
        "Connecticut",
        "Delaware",
        "Florida",
        "Georgia",
        "Hawaii",
        "Idaho",
        "Illinois",
        "Indiana",
        "Iowa",
        "Kansas",
        "Kentucky",
        "Louisiana",
        "Maine",
        "Maryland",
        "Massachusetts",
        "Michigan",
        "Minnesota",
        "Mississippi",
        "Missouri",
        "Montana",
        "Nebraska",
        "Nevada",
        "New Hampshire",
        "New Jersey",
        "New Mexico",
        "New York",
        "North Carolina",
        "North Dakota",
        "Ohio",
        "Oklahoma",
        "Oregon",
        "Pennsylvania",
        "Rhode Island",
        "South Carolina",
        "South Dakota",
        "Tennessee",
        "Texas",
        "Utah",
        "Vermont",
        "Virginia",
        "Washington",
        "West Virginia",
        "Wisconsin",
        "Wyoming",
    )

    def mount(self):
        self.flavors = Flavor.objects.all()[:3]

    def set_name(self, name=None):
        if name:
            self.name = name
        else:
            self.name = "Universe"

    def clear_states(self):
        self.state = ""

    def states(self):
        if not self.state:
            return []

        return [s for s in self.ALL_STATES if s.lower().startswith(self.state.lower())]

    class Meta:
        exclude = ("ALL_STATES",)



import logging
import re
from collections import deque

import orjson
from bs4 import BeautifulSoup
from bs4.dammit import EntitySubstitution
from bs4.element import Tag
from bs4.formatter import HTMLFormatter
from django.template.backends.django import Template
from django.template.response import TemplateResponse

from django_unicorn.decorators import timed
from django_unicorn.errors import (
    MissingComponentElementError,
    MissingComponentViewElementError,
    MultipleRootComponentElementError,
    NoRootComponentElementError,
)
from django_unicorn.settings import get_minify_html_enabled, get_script_location
from django_unicorn.utils import generate_checksum, sanitize_html

logger = logging.getLogger(__name__)


# https://developer.mozilla.org/en-US/docs/Glossary/Empty_element
EMPTY_ELEMENTS = (
    "<area>",
    "<base>",
    "<br>",
    "<col>",
    "<embed>",
    "<hr>",
    "<img>",
    "<input>",
    "<link>",
    "<meta>",
    "<param>",
    "<source>",
    "<track>",
    "<wbr>",
)


def is_html_well_formed(html: str) -> bool:
    """
    Whether the passed-in HTML is missing any closing elements which can cause issues when rendering.
    """

    tag_list = re.split("(<[^>!]*>)", html)[1::2]
    stack: deque[str] = deque()

    for tag in tag_list:
        if "/" not in tag:
            cleaned_tag = re.sub(r"(<([\w\-]+)[^>!]*>)", r"<\2>", tag)

            if cleaned_tag not in EMPTY_ELEMENTS:
                stack.append(cleaned_tag)
        elif len(stack) > 0 and (tag.replace("/", "") == stack[len(stack) - 1]):
            stack.pop()

    return len(stack) == 0


def assert_has_single_wrapper_element(root_element: Tag, component_name: str) -> None:
    """Assert that there is at least one child in the root element. And that there is only
    one root element.
    """

    # Check that the root element has at least one child
    try:
        next(root_element.descendants)
    except StopIteration:
        raise NoRootComponentElementError(
            f"The '{component_name}' component does not appear to have one root element."
        ) from None

    if "unicorn:view" in root_element.attrs or "u:view" in root_element.attrs:
        # If the root element is a direct view, skip the check
        return

    # Check that there is not more than one root element
    parent_element = root_element.parent
    if parent_element is None:
        raise AssertionError("Root element must have a parent")

    tag_count = len([c for c in parent_element.children if isinstance(c, Tag)])

    if tag_count > 1:
        raise MultipleRootComponentElementError(
            f"The '{component_name}' component appears to have multiple root elements."
        ) from None


def _get_direct_view(tag: Tag):
    return tag.find_next(attrs={"unicorn:view": True}) or tag.find_next(attrs={"u:view": True})


def get_root_element(soup: BeautifulSoup) -> Tag:
    """Gets the first tag element for the component or the first element with a `unicorn:view` attribute for a direct
    view.

    Returns:
        BeautifulSoup tag element.

        Raises `Exception` if an element cannot be found.
    """

    for element in soup.contents:
        if isinstance(element, Tag) and element.name:
            if element.name == "html":
                view_element = _get_direct_view(element)

                if not view_element:
                    raise MissingComponentViewElementError(
                        "An element with an `unicorn:view` attribute is required for a direct view"
                    )

                return view_element

            return element

    raise MissingComponentElementError("No root element for the component was found")


class UnsortedAttributes(HTMLFormatter):
    """
    Prevent beautifulsoup from re-ordering attributes.
    """

    def __init__(self):
        super().__init__(entity_substitution=EntitySubstitution.substitute_html)

    def attributes(self, tag: Tag):
        yield from tag.attrs.items()


class UnicornTemplateResponse(TemplateResponse):
    def __init__(
        self,
        template,
        request,
        *,
        context=None,
        content_type=None,
        status=None,
        charset=None,
        using=None,
        component=None,
        init_js=False,
        **kwargs,  # noqa: ARG002
    ):
        super().__init__(
            template=template,
            request=request,
            context=context,
            content_type=content_type,
            status=status,
            charset=charset,
            using=using,
        )

        self.component = component
        self.init_js = init_js

    def resolve_template(self, template):
        """Override the TemplateResponseMixin to resolve a list of Templates.

        Calls the super which accepts a template object, path-to-template, or list of paths if the first
        object in the sequence is not a Template.
        """

        if isinstance(template, list | tuple):
            if isinstance(template[0], Template):
                return template[0]

        return super().resolve_template(template)

    @timed
    def render(self):
        response = super().render()

        if not self.component or not self.component.component_id:
            return response

        content = response.content.decode("utf-8")

        if not is_html_well_formed(content):
            logger.warning(
                f"The HTML in '{self.component.component_name}' appears to be missing a closing tag. That can \
potentially cause errors in Unicorn."
            )

        frontend_context_variables = self.component.get_frontend_context_variables()
        frontend_context_variables_dict = orjson.loads(frontend_context_variables)
        checksum = generate_checksum(frontend_context_variables_dict)

        # Use `html.parser` and not `lxml` because in testing it was no faster even with `cchardet`
        # despite https://thehftguy.com/2020/07/28/making-beautifulsoup-parsing-10-times-faster/
        soup = BeautifulSoup(content, features="html.parser")
        root_element = get_root_element(soup)

        try:
            assert_has_single_wrapper_element(root_element, self.component.component_name)
        except (NoRootComponentElementError, MultipleRootComponentElementError) as ex:
            logger.warning(ex)

        root_element["unicorn:id"] = self.component.component_id
        root_element["unicorn:name"] = self.component.component_name
        root_element["unicorn:key"] = self.component.component_key
        root_element["unicorn:checksum"] = checksum
        root_element["unicorn:data"] = frontend_context_variables
        root_element["unicorn:calls"] = orjson.dumps(self.component.calls).decode("utf-8")

        # Generate the checksum based on the rendered content (without script tag)
        content_hash = generate_checksum(UnicornTemplateResponse._desoupify(soup))

        if self.init_js:
            init = {
                "id": self.component.component_id,
                "name": self.component.component_name,
                "key": self.component.component_key,
                "data": orjson.loads(frontend_context_variables),
                "calls": self.component.calls,
                "hash": content_hash,
            }
            init = orjson.dumps(init).decode("utf-8")
            json_element_id = f"unicorn:data:{self.component.component_id}"
            init_script = (
                f"Unicorn.componentInit(JSON.parse(document.getElementById('{json_element_id}').textContent));"
            )

            json_tag = soup.new_tag("script")
            json_tag["type"] = "application/json"
            json_tag["id"] = json_element_id
            json_tag.string = sanitize_html(init)

            if self.component.parent:
                self.component._init_script = init_script
                self.component._json_tag = json_tag
            else:
                json_tags = [json_tag]

                descendants = []
                descendants.append(self.component)
                while descendants:
                    descendant = descendants.pop()
                    for child in descendant.children:
                        init_script = f"{init_script} {child._init_script}"
                        json_tags.append(child._json_tag)

                        # We need to delete this property here as it can cause RecursionError
                        # when pickling child component. Tag element has previous_sibling
                        # and next_sibling which would also be pickled and if they are big,
                        # cause RecursionError
                        del child._json_tag

                        descendants.append(child)

                script_tag = soup.new_tag("script")
                script_tag["type"] = "module"
                script_tag.string = f"if (typeof Unicorn === 'undefined') {{ console.error('Unicorn is missing. Do you \
need {{% load unicorn %}} or {{% unicorn_scripts %}}?') }} else {{ {init_script} }}"

                if get_script_location() == "append":
                    root_element.append(script_tag)
                else:
                    root_element.insert_after(script_tag)

                for t in json_tags:
                    if get_script_location() == "append":
                        root_element.append(t)
                    else:
                        root_element.insert_after(t)

        rendered_template = UnicornTemplateResponse._desoupify(soup)
        self.component.rendered(rendered_template)

        response.content = rendered_template

        if get_minify_html_enabled():
            # Import here in case the minify extra was not installed
            from htmlmin import minify  # type: ignore # noqa: PLC0415

            minified_html = minify(response.content.decode())

            if len(minified_html) < len(rendered_template):
                response.content = minified_html

        return response

    @staticmethod
    def _desoupify(soup):
        soup.smooth()
        return soup.encode(formatter=UnsortedAttributes()).decode("utf-8")



import os
import webbrowser
from pathlib import Path

from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError

from django_unicorn.components.unicorn_view import (
    convert_to_pascal_case,
    convert_to_snake_case,
)

COMPONENT_FILE_CONTENT = """from django_unicorn.components import UnicornView


class {pascal_case_component_name}View(UnicornView):
    pass
"""

TEMPLATE_FILE_CONTENT = """<div>
    <!-- put component code here -->
</div>
"""


def get_app_path(app_name: str) -> Path:
    """
    Gets the directory path for an installed application.
    """

    return Path(apps.get_app_config(app_name).path)


class Command(BaseCommand):
    help = "Creates a new component for `django-unicorn`"

    def add_arguments(self, parser):
        parser.add_argument("app_name", type=str)
        parser.add_argument("component_names", nargs="+", type=str, help="Names of components")

    def check_initials_directories(self, app_directory: Path) -> tuple[dict[str, Path], bool]:
        """
        Checks for directories existance and creates them if necessary.
        Returns a tuple containing a dictonary `components` and `templates`
        keys, and a boolean indicating if directories were created.
        """

        paths = {
            "components": app_directory / "components",
            "templates": app_directory / "templates" / "unicorn",
        }

        is_first = False

        if not paths["components"].exists():
            paths["components"].mkdir()
            self.stdout.write(self.style.SUCCESS(f"Created your first component in '{app_directory.name}'! ✨\n"))
            is_first = True

        (paths["components"] / "__init__.py").touch(exist_ok=True)

        if not paths["templates"].exists():
            paths["templates"].mkdir(parents=True, exist_ok=True)
            self.stdout.write(self.style.SUCCESS(f"Created your first template in '{app_directory.name}'! ✨\n"))
            is_first = True

        return paths, is_first

    def obtain_nested_path(self, component_name: str) -> tuple[str, str]:
        """
        Receives the complete component name and returns a tuple
        with the nested path and component name itself.
        """

        nested_paths = []

        if "." in component_name:
            (*nested_paths, component_name) = component_name.split(".")

        return "/".join(nested_paths), component_name

    def create_nested_directories(self, paths: dict[str, Path], nested_path: str) -> None:
        """
        Creates the nested directories for the components and templates.
        """

        nested_paths = nested_path.split("/")

        component_path = paths["components"]
        template_path = paths["templates"]

        for _nested_path in nested_paths:
            component_path /= _nested_path
            template_path /= _nested_path

            if not component_path.exists():
                component_path.mkdir()

            if not template_path.exists():
                template_path.mkdir()

            (component_path / "__init__.py").touch(exist_ok=True)

    def create_component_and_template(self, paths: dict[str, Path], nested_path: str, component_name: str) -> None:
        """
        Creates the component and template files.
        """

        snake_case_component_name = convert_to_snake_case(component_name)
        pascal_case_component_name = convert_to_pascal_case(component_name)

        component_path = paths["components"] / nested_path / f"{snake_case_component_name}.py"

        if component_path.exists():
            self.stdout.write(
                self.style.ERROR(f"Skipping creating {snake_case_component_name}.py because it already exists.")
            )
        else:
            component_path.write_text(
                COMPONENT_FILE_CONTENT.format(**{"pascal_case_component_name": pascal_case_component_name})
            )
            self.stdout.write(self.style.SUCCESS(f"Created {component_path}."))

        template_path = paths["templates"] / nested_path / f"{component_name}.html"

        if template_path.exists():
            self.stdout.write(self.style.ERROR(f"Skipping creating {component_name}.html because it already exists."))
        else:
            template_path.write_text(TEMPLATE_FILE_CONTENT)
            self.stdout.write(self.style.SUCCESS(f"Created {template_path}."))

    def ask_for_star_repo(self) -> None:
        will_star_repo = input(
            "\nStarring the GitHub repo helps other Django users find Unicorn. Can you star it for me? [y/N] "
        )

        if will_star_repo.strip().lower() in ("y", "yes"):
            self.stdout.write(self.style.SUCCESS("Thank you for helping spread the word about Unicorn!"))

            self.stdout.write(
                """
                         ,/
                        //
                      ,//
          __    /|   |//
      `__/\\_ --(/|___/-/
   \\|\\_-\\___ __-_`- /-/ \\.
  |\\_-___,-\\_____--/_)' ) \\
   \\ -_ /     __ \\( `( __`\\|
   `\\__|      |\\)      (/|
',--//-|      \\    | '   /
  /,---|       \\        /
 `_/ _,'        |      |
 __/'/          |      |
 ___/           \\ (  ) /
                 \\____/\\
                        \\
"""
            )

            webbrowser.open("https://github.com/adamghill/django-unicorn", new=2)
        else:
            self.stdout.write(
                self.style.ERROR("That's a bummer, but I understand. I hope you will star it for me later!")
            )

    def handle(self, **options):
        # Default from `django-cookiecutter`
        base_path = getattr(settings, "APPS_DIR", None)

        if not base_path:
            # Default from new Django project
            base_path = getattr(settings, "BASE_DIR", None)

        if not base_path:
            # Fallback to the current directory
            base_path = os.getcwd()

        base_path = Path(base_path)

        if "app_name" not in options:
            raise CommandError("An application name is required.")

        if "component_names" not in options:
            raise CommandError("At least one component name is required.")

        app_name = options["app_name"]
        is_new_app = False

        try:
            app_directory = get_app_path(app_name)
        except LookupError as e:
            should_create_app = input(
                f"\n'{app_name}' cannot be found. Should it be created automatically with `startapp {app_name}`? [y/N] "
            )

            if should_create_app.strip().lower() in ("y", "yes"):
                call_command(
                    "startapp",
                    app_name,
                    verbosity=0,
                )
                app_directory = base_path / app_name

                is_new_app = True
            else:
                raise CommandError(
                    f"An app named '{app_name}' does not exist yet. You might need to create it first."
                ) from e

        if not app_directory.exists():
            app_directory.mkdir()

        paths, is_first = self.check_initials_directories(app_directory)

        for component_name in options["component_names"]:
            (nested_path, name) = self.obtain_nested_path(component_name)

            self.create_nested_directories(paths, nested_path)

            self.create_component_and_template(paths, nested_path, name)

            if is_first:
                self.ask_for_star_repo()
                is_first = False

            if is_new_app:
                msg = f'\nMake sure to add `"{app_name}",` to your INSTALLED_APPS list in your settings file if necessary.'  # noqa: E501
                self.stdout.write(self.style.WARNING(msg))



import sys
from datetime import datetime, timezone
from typing import List  # noqa: UP035

import pytest
from django.db.models import Model, QuerySet

from django_unicorn.components import UnicornView
from django_unicorn.typing import QuerySetType
from django_unicorn.views.utils import set_property_from_data
from example.coffee.models import Flavor


class FakeComponent(UnicornView):
    string = "property_view"
    integer = 99
    datetime_without_typehint = datetime(2020, 1, 1, tzinfo=timezone.utc)
    datetime_with_typehint: datetime = datetime(2020, 2, 1, tzinfo=timezone.utc)
    array: list[str] = []  # noqa: RUF012
    model = Flavor(name="test-initial")
    queryset = Flavor.objects.none()
    queryset_with_typehint: QuerySetType[Flavor] = []  # noqa: RUF012
    datetimes = [datetime(2020, 3, 1, tzinfo=timezone.utc)]  # noqa: RUF012
    datetimes_with_old_typehint: List[datetime] = [datetime(2020, 4, 1, tzinfo=timezone.utc)]  # noqa: RUF012, UP006
    datetimes_with_list_typehint: list = [datetime(2020, 6, 1, tzinfo=timezone.utc)]  # noqa: RUF012

    try:
        datetimes_with_new_typehint: list[datetime] = [datetime(2020, 5, 1, tzinfo=timezone.utc)]
    except TypeError:
        datetimes_with_new_typehint: None


class FakeQuerySetComponent(UnicornView):
    queryset_with_typehint: QuerySetType[Flavor] = None


class FakeDbComponent(UnicornView):
    queryset_with_data = Flavor.objects.none()

    def __init__(self, **kwargs):
        flavor = Flavor(pk=1, name="initial-test-data")
        flavor.save()
        self.queryset_with_data = Flavor.objects.filter()[:1]
        super().__init__(**kwargs)


class FakeAllQuerySetComponent(UnicornView):
    queryset_with_empty_list: QuerySetType[Flavor] = []  # noqa: RUF012
    queryset_with_none: QuerySetType[Flavor] = None
    queryset_with_empty_queryset: QuerySetType[Flavor] = Flavor.objects.none()
    queryset_with_no_typehint = Flavor.objects.none()


def component_queryset_field_asserts(component, field_name):
    field = getattr(component, field_name)
    assert isinstance(field, QuerySet)
    assert field.model == Flavor
    assert len(field) == 1
    assert field[0].name == "test-qs"


def test_set_property_from_data_str():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_str")
    assert "property_view" == component.string

    set_property_from_data(component, "string", "property_view_updated")

    assert "property_view_updated" == component.string


def test_set_property_from_data_int():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_int")
    assert 99 == component.integer

    set_property_from_data(component, "integer", 100)

    assert 100 == component.integer


def test_set_property_from_data_datetime():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_datetime")
    assert datetime(2020, 1, 1, tzinfo=timezone.utc) == component.datetime_without_typehint

    set_property_from_data(
        component,
        "datetime_without_typehint",
        datetime(2020, 1, 2, tzinfo=timezone.utc),
    )

    assert datetime(2020, 1, 2, tzinfo=timezone.utc) == component.datetime_without_typehint


def test_set_property_from_data_datetime_with_typehint():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_datetime_with_typehint")
    assert datetime(2020, 2, 1, tzinfo=timezone.utc) == component.datetime_with_typehint

    set_property_from_data(
        component,
        "datetime_with_typehint",
        str(datetime(2020, 2, 2, tzinfo=timezone.utc)),
    )

    assert datetime(2020, 2, 2, tzinfo=timezone.utc) == component.datetime_with_typehint


def test_set_property_from_data_list():
    """
    Prevent attempting to instantiate `List[]` type-hint doesn't throw TypeError
    """
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_list")
    assert component.array == []

    set_property_from_data(component, "array", ["string"])

    assert ["string"] == component.array


def test_set_property_from_data_list_datetimes():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_list_datetimes")
    assert [datetime(2020, 3, 1, tzinfo=timezone.utc)] == component.datetimes

    set_property_from_data(component, "datetimes", [str(datetime(2020, 3, 2, tzinfo=timezone.utc))])

    assert [str(datetime(2020, 3, 2, tzinfo=timezone.utc))] == component.datetimes


def test_set_property_from_data_list_datetimes_with_old_typehint():
    component = FakeComponent(
        component_name="test", component_id="test_set_property_from_data_list_datetimes_with_old_typehint"
    )
    assert [datetime(2020, 4, 1, tzinfo=timezone.utc)] == component.datetimes_with_old_typehint

    set_property_from_data(
        component,
        "datetimes_with_old_typehint",
        [str(datetime(2020, 4, 2, tzinfo=timezone.utc))],
    )

    assert [datetime(2020, 4, 2, tzinfo=timezone.utc)] == component.datetimes_with_old_typehint


@pytest.mark.skipif(
    sys.version_info.major == 3 and sys.version_info.minor <= 8,  # noqa: YTT204
    reason="Skip new type hints for Python 3.8 or less",
)
def test_set_property_from_data_list_datetimes_with_new_typehint():
    component = FakeComponent(
        component_name="test", component_id="test_set_property_from_data_list_datetimes_with_new_typehint"
    )
    assert [datetime(2020, 5, 1, tzinfo=timezone.utc)] == component.datetimes_with_new_typehint

    set_property_from_data(
        component,
        "datetimes_with_new_typehint",
        [str(datetime(2020, 5, 2, tzinfo=timezone.utc))],
    )

    assert [datetime(2020, 5, 2, tzinfo=timezone.utc)] == component.datetimes_with_new_typehint


def test_set_property_from_data_list_datetimes_with_list_typehint():
    component = FakeComponent(
        component_name="test", component_id="test_set_property_from_data_list_datetimes_with_list_typehint"
    )
    assert [datetime(2020, 6, 1, tzinfo=timezone.utc)] == component.datetimes_with_list_typehint

    set_property_from_data(
        component,
        "datetimes_with_list_typehint",
        [str(datetime(2020, 6, 2, tzinfo=timezone.utc))],
    )

    assert [str(datetime(2020, 6, 2, tzinfo=timezone.utc))] == component.datetimes_with_list_typehint


def test_set_property_from_data_model():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_model")
    assert component.model.name == "test-initial"

    set_property_from_data(component, "model", {"name": "test-test"})

    assert isinstance(component.model, Model)
    assert isinstance(component.model, Flavor)
    assert component.model.name == "test-test"


def test_set_property_from_data_empty_queryset():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_empty_queryset")
    assert len(component.queryset) == 0

    set_property_from_data(component, "queryset", [{"name": "test-qs"}])

    component_queryset_field_asserts(component, "queryset")


@pytest.mark.django_db
def test_set_property_from_data_queryset():
    component = FakeDbComponent(component_name="test", component_id="test_set_property_from_data_queryset")
    assert len(component.queryset_with_data) == 1

    set_property_from_data(component, "queryset_with_data", [{"pk": 1, "name": "test-qs"}])

    component_queryset_field_asserts(component, "queryset_with_data")


def test_set_property_from_data_queryset_list_with_typehint():
    component = FakeComponent(
        component_name="test", component_id="test_set_property_from_data_queryset_list_with_typehint"
    )
    assert len(component.queryset_with_typehint) == 0

    set_property_from_data(component, "queryset_with_typehint", [{"name": "test-qs"}])

    component_queryset_field_asserts(component, "queryset_with_typehint")


def test_set_property_from_data_queryset_none_with_typehint():
    component = FakeQuerySetComponent(
        component_name="test", component_id="test_set_property_from_data_queryset_none_with_typehint"
    )
    assert component.queryset_with_typehint is None

    set_property_from_data(component, "queryset_with_typehint", [{"name": "test-qs"}])

    component_queryset_field_asserts(component, "queryset_with_typehint")


def test_set_property_from_data_queryset_parent():
    component = FakeQuerySetComponent(component_name="test", component_id="test_set_property_from_data_queryset_parent")
    assert component.queryset_with_typehint is None

    set_property_from_data(component, "queryset_with_typehint", [{"name": "test-qs"}])

    component_queryset_field_asserts(component, "queryset_with_typehint")


def test_set_property_from_data_all_querysets():
    component = FakeAllQuerySetComponent(
        component_name="test", component_id="test_set_property_from_data_all_querysets"
    )

    set_property_from_data(component, "queryset_with_empty_list", [{"name": "test-qs"}])
    set_property_from_data(component, "queryset_with_none", [{"name": "test-qs"}])
    set_property_from_data(component, "queryset_with_empty_queryset", [{"name": "test-qs"}])
    set_property_from_data(component, "queryset_with_no_typehint", [{"name": "test-qs"}])

    component_queryset_field_asserts(component, "queryset_with_empty_list")
    component_queryset_field_asserts(component, "queryset_with_none")
    component_queryset_field_asserts(component, "queryset_with_empty_queryset")
    component_queryset_field_asserts(component, "queryset_with_no_typehint")


@pytest.mark.django_db
def test_set_property_from_data_many_to_many():
    component = FakeComponent(component_name="test", component_id="test_set_property_from_data_many_to_many")
    component.model.pk = 1

    # No `TypeError: Direct assignment to the reverse side of a many-to-many set is prohibited.` error gets raised
    set_property_from_data(component.model, "taste_set", [])



import shortuuid
from tests.views.fake_components import FakeComponent
from tests.views.message.test_calls import FakeCallsComponent
from tests.views.message.utils import post_and_get_response

from django_unicorn.components import UnicornView
from django_unicorn.utils import generate_checksum


class FakeComponentParent(UnicornView):
    template_name = "templates/test_component_parent.html"

    value: int = 0


class FakeComponentParentWithValue(UnicornView):
    template_name = "templates/test_component_parent_with_value.html"

    value: int = 0


class FakeComponentChild(UnicornView):
    template_name = "templates/test_component_child.html"

    def parent_increment(self):
        self.parent.value += 1


def test_message_hash_no_change(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponent(
        component_id=component_id,
        component_name="tests.views.fake_components.FakeComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {"method_count": 0}
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_method_kwargs(count=0)"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    assert response.status_code == 304


def test_message_hash_changes(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponent(
        component_id=component_id,
        component_name="tests.views.fake_components.FakeComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {"method_count": 0}
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_method_kwargs(count=1)"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    assert response["data"]["method_count"] == 1


def test_message_hash_no_change_but_return_value(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponent(
        component_id=component_id,
        component_name="tests.views.fake_components.FakeComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {"method_count": 0}
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_return_value"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    # check that the response is JSON and not a 304
    assert isinstance(response, dict)
    assert response["return"]["value"]


def test_message_hash_no_change_but_return_redirect(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponent(
        component_id=component_id,
        component_name="tests.views.fake_components.FakeComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {"method_count": 0}
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_redirect"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    # check that the response is JSON and not a 304
    assert isinstance(response, dict)
    assert response["return"]["value"]


def test_message_hash_no_change_but_return_hash_update(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponent(
        component_id=component_id,
        component_name="tests.views.fake_components.FakeComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {"method_count": 0}
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_hash_update"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    # check that the response is JSON and not a 304
    assert isinstance(response, dict)
    assert response["return"]["value"]


def test_message_hash_no_change_but_return_poll_update(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponent(
        component_id=component_id,
        component_name="tests.views.fake_components.FakeComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {"method_count": 0}
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_poll_update"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    # check that the response is JSON and not a 304
    assert isinstance(response, dict)
    assert response["return"]["value"]


def test_message_hash_no_change_but_return_location_update(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponent(
        component_id=component_id,
        component_name="tests.views.fake_components.FakeComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {"method_count": 0}
    response = post_and_get_response(
        client,
        url="/message/tests.views.fake_components.FakeComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_refresh_redirect"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    # check that the response is JSON and not a 304
    assert isinstance(response, dict)
    assert response["return"]["value"]


def test_message_hash_no_change_but_calls(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeCallsComponent(
        component_id=component_id,
        component_name="tests.views.message.test_calls.FakeCallsComponent",
    )
    rendered_content = component.render()
    checksum = generate_checksum(rendered_content)

    data = {}
    response = post_and_get_response(
        client,
        url="/message/tests.views.message.test_calls.FakeCallsComponent",
        data=data,
        action_queue=[
            {
                "payload": {"name": "test_call"},
                "type": "callMethod",
            }
        ],
        component_id=component_id,
        hash=checksum,
    )

    # check that the response is JSON and not a 304
    assert isinstance(response, dict)
    assert response.get("calls") == [{"args": [], "fn": "testCall"}]


def test_message_hash_no_change_but_parent(client):
    component_id = shortuuid.uuid()[:8]
    component = FakeComponentParentWithValue(
        component_id=component_id,
        component_name="tests.views.message.test_hash.FakeComponentParentWithValue",
    )
    component.render()

    assert component.value == 0

    child = component.children[0]
    rendered_child_content = child.render()
    child_hash = generate_checksum(rendered_child_content)

    assert child.parent.value == 0

    data = {}
    response = post_and_get_response(
        client,
        url="/message/tests.views.message.test_hash.FakeComponentChild",
        data=data,
        action_queue=[
            {
                "payload": {"name": "parent_increment()"},
                "type": "callMethod",
            }
        ],
        component_id=child.component_id,
        hash=child_hash,
        return_response=True,
    )

    assert response.status_code == 200

    # TODO: Revisit this assert after child upgrade
    # assert child.parent.value == 1

    # rendered_parent_content = child.parent.render()
    # assert "||value:1||" in rendered_parent_content
